"""
This file is part of pyspex

https://github.com/rmvanhees/pyspex.git

Python class to create SPEXone Level-1 products

Copyright (c) 2019-2021 SRON - Netherlands Institute for Space Research
   All Rights Reserved

License:  BSD-3-Clause
"""
from datetime import datetime, timedelta, timezone
from pathlib import Path, PurePosixPath

import numpy as np

from netCDF4 import Dataset

from .tm_science import TMscience
from .lib.attrs_def import attrs_def
from .lib.l1a_def import init_l1a
from .lib.l1b_def import init_l1b
from .lib.l1c_def import init_l1c

# - global parameters -------------------
MCP_TO_SEC = 1e-7


def frac_poly(xx_in, coefs=None):
    """
    Temperature [K] calibration derived by Paul Tol (2020-10-21)

    Parameters
    ----------
    xx    :  ndarray
    coefs :  tuple, default=None
      coefficients of fractional polynomial: r0, r1, r2, r3, r4

    Returns
    -------
    ndarray, dtype float
    """
    xx = xx_in.astype(float)

    if coefs is None:
        coefs = (273.15 + 21.19, 6.97828e+7,
                 -3.53275e-25, 7.79625e-31, -4.6505E-32)

    return (coefs[0]
            + coefs[1] / xx
            + coefs[2] * xx ** 4
            + (coefs[3] + coefs[4] * np.log(xx)) * xx ** 5)


# - class LV1io -------------------------
class Lv1io:
    """
    Generic class to create SPEXone Level-1 products

    Attributes
    ----------
    product: pathlib.Path object
       Concrete path object to SPEXone Level-1 product
    fid: netCDF5.Dataset object
       NetCDF4 Pointer to SPEXone Level-1 product
    dset_stored: dict
       Number of items stored for all required netCDF4 variables

    Methods
    -------
    close()
       Close all resources (currently a placeholder function).
    epoch
       Provide epoch for SPEXone.
    get_dim(ds_name)
       Returns size of dimension.
    get_attr(attr_name, ds_name=None)
       Read data of an attribute, global or attached to a group or variable.
    set_attr(attr_name, value, ds_name=None)
       Write data to an attribute, global or attached to a group or variable.
    get_dset(ds_name)
       Read data of a netCDF4 variable.
    set_dset(ds_name, value, ibgn=-1)
       Write/append data to a netCDF4 variable.
    fill_global_attrs(level, orbit=-1, bin_size=None, inflight=False)
       Define global attributes in the SPEXone Level-1 products.

    Notes
    -----
    The engineering data should be extended, suggestions:
    * Temperatures of a.o. detector, FEE, optica, obm, telescope
    * Instrument settings: exposure time, dead time, frame time, coadding, ...
    """
    processing_level = 'unknown'
    dset_stored = {}

    def __init__(self, product: str, append=False, **kwargs):
        """
        Initialize access to a SPEXone Level-1 product

        Parameters
        ----------
        product : str
           name of the SPEXone Level-1 product
        append : bool, default=False
           do no clobber, but add new data to existing product
        """
        self.product = Path(product)
        self.fid = None

        # initialize private class-attributes
        self.__epoch = datetime(1970, 1, 1, tzinfo=timezone.utc)

        # initialize Level-1 product
        if not append:
            if self.processing_level == 'L1A':
                self.fid = init_l1a(product, **kwargs)
            elif self.processing_level == 'L1B':
                self.fid = init_l1b(product, **kwargs)
                if 'ref_date' in kwargs and kwargs['ref_date'] is not None:
                    self.__epoch = kwargs['ref_date']
            elif self.processing_level == 'L1C':
                self.fid = init_l1c(product, **kwargs)
                if 'ref_date' in kwargs and kwargs['ref_date'] is not None:
                    self.__epoch = kwargs['ref_date']
            else:
                raise KeyError('valid processing levels are: L1A, L1B or L1C')
        else:
            # open Level-1 product in append mode
            self.fid = Dataset(self.product, "r+")

            # store current length of the first dimension
            for key in self.dset_stored:
                self.dset_stored[key] = self.fid[key].shape[0]

    def __repr__(self) -> str:
        class_name = type(self).__name__
        return f'{class_name}({self.product!r})'

    def __iter__(self):
        for attr in sorted(self.__dict__):
            if not attr.startswith("__"):
                yield attr

    def __enter__(self):
        """
        method called to initiate the context manager
        """
        return self

    def __exit__(self, exc_type, exc_value, traceback) -> bool:
        """
        method called when exiting the context manager
        """
        self.close()
        return False  # any exception is raised by the with statement.

    def close(self) -> None:
        """
        Close all resources (currently a placeholder function)
        """
        if self.fid is None:
            return

        self.fid.close()
        self.fid = None

    # ---------- PUBLIC FUNCTIONS ----------
    @property
    def epoch(self) -> datetime:
        """
        Provide epoch for SPEXone
        """
        return self.__epoch

    def get_dim(self, name: str):
        """
        Get size of a netCDF4 dimension
        """
        return self.fid.dimensions[name].size

    # ----- ATTRIBUTES --------------------
    def get_attr(self, name: str, ds_name=None):
        """
        Read data of an attribute, global or attached to a group or variable

        Parameters
        ----------
        name : string
           name of the attribute
        ds_name : string, default=None
           name of dataset to which the attribute is attached

        Returns
        -------
        scalar or array_like
           value of attribute 'name', global or attached to dataset 'ds_name'
        """
        if ds_name is None:
            res = self.fid.getncattr(name)
        else:
            if ds_name not in self.fid.groups \
               and ds_name not in self.fid.variables:
                return None
            res = self.fid[ds_name].getncattr(name)

        if isinstance(res, bytes):
            return res.decode('ascii')

        return res

    def set_attr(self, name: str, value, ds_name=None) -> None:
        """
        Write data to an attribute, global or attached to a group or variable

        Parameters
        ----------
        name : string
           name of the attribute
        value : scalar, array_like
           value or values to be written
        ds_name : string, default=None
           name of group or dataset to which the attribute is attached
           **Use group name without starting '/'**
        """
        if ds_name is None:
            if isinstance(value, str):
                self.fid.setncattr(name, np.string_(value))
            else:
                self.fid.setncattr(name, value)
        else:
            grp_name = str(PurePosixPath(ds_name).parent)
            var_name = str(PurePosixPath(ds_name).name)
            if grp_name != '.':
                if var_name not in self.fid[grp_name].groups \
                   and var_name not in self.fid[grp_name].variables:
                    raise KeyError(f'ds_name {ds_name} not present in product')
            else:
                if var_name not in self.fid.groups \
                   and var_name not in self.fid.variables:
                    raise KeyError(f'ds_name {ds_name} not present in product')

            if isinstance(value, str):
                self.fid[ds_name].setncattr(name, np.string_(value))
            else:
                self.fid[ds_name].setncattr(name, value)

    # ----- VARIABLES --------------------
    def get_dset(self, name: str):
        """
        Read data of a netCDF4 variable

        Parameters
        ----------
        name : string
           name of dataset

        Returns
        -------
        scalar or array_like
           value of dataset 'name'
        """
        grp_name = str(PurePosixPath(name).parent)
        var_name = str(PurePosixPath(name).name)
        if grp_name != '.':
            if var_name not in self.fid[grp_name].variables:
                raise KeyError(f'dataset {name} not present in Level-1 product')
        else:
            if var_name not in self.fid.variables:
                raise KeyError(f'dataset {name} not present in Level-1 product')

        return self.fid[name][:]

    def set_dset(self, name: str, value, ibgn=-1) -> None:
        """
        Write/append data to a netCDF4 variable

        Parameters
        ----------
        name : string
           Name of Level-1 dataset
        value : scalar or array_like
           Value or values to be written
        ibgn : int, default=-1
           Index of the first (unlimited) dimension where to store the new data
           Default is to append the data
        """
        value = np.asarray(value)
        grp_name = str(PurePosixPath(name).parent)
        var_name = str(PurePosixPath(name).name)
        if grp_name != '.':
            if var_name not in self.fid[grp_name].variables:
                raise KeyError(f'dataset {name} not present in Level-1 product')
        else:
            if var_name not in self.fid.variables:
                raise KeyError(f'dataset {name} not present in Level-1 product')

        dims = self.fid[name].get_dims()
        if not dims:
            self.fid[name][...] = value
        elif dims[0].isunlimited():
            if ibgn < 0:
                ibgn = self.dset_stored[name]
            self.fid[name][ibgn:, ...] = value
        else:
            self.fid[name][...] = value

        self.dset_stored[name] += 1 if value.shape == () else value.shape[0]

    # -------------------------
    def fill_global_attrs(self, orbit=-1,
                          bin_size=None,
                          inflight=False) -> None:
        """
        Define global attributes in the SPEXone Level-1 products

        Parameters
        ----------
        orbit_number :  int, default=-1
           Orbit revolution counter
        bin_size :  str, default=None
           Size of the nadir footprint (cross-track), include unit: e.g. '5km'
        inflight :  bool, default=False
           Measurements performed on-ground or inflight
        """
        dict_attrs = attrs_def(self.processing_level, inflight)
        dict_attrs['product_name'] = self.product.name
        dict_attrs['orbit_number'] = orbit
        if bin_size is not None:
            dict_attrs['bin_size_at_nadir'] = bin_size

        for key, value in dict_attrs.items():
            if value is not None:
                self.fid.setncattr(key, value)


# - class L1Aio -------------------------
class L1Aio(Lv1io):
    """
    This class can be used to create a SPEXone Level-1A product

    Parameters
    ----------
    lv1_product: string
       Name of the Level-1A product
    append : boolean, default=False
       If the file is opened in append mode, then parameter 'dims' is ignored
    dims: dictionary, default=None
       Provide size of various dimensions (L1A only).
       Default values:
            number_of_images : None     # number of image frames
            samples_per_image : 184000  # depends on binning table
            SC_records : None           # space-craft navigation records
            hk_packets : None           # number of HK tlm-packets
            wavelength : None

    Attributes
    ----------
    product: pathlib.Path object
       Concrete path object to SPEXone Level-1 product
    fid: netCDF5.Dataset object
       NetCDF4 Pointer to SPEXone Level-1 product
    dset_stored: dict
       Number of items stored for all required netCDF4 variables

    Methods
    -------
    close()
       Close product and check if required datasets are filled with data.
    epoch
       Provide epoch for SPEXone.
    get_dim(ds_name)
       Returns size of dimension.
    get_attr(attr_name, ds_name=None)
       Read data of an attribute, global or attached to a group or variable.
    set_attr(attr_name, value, ds_name=None)
       Write data to an attribute, global or attached to a group or variable.
    get_dset(ds_name)
       Read data of a netCDF4 variable.
    set_dset(ds_name, valu, ibgn=-1e)
       Write/append data to a netCDF4 variable.
    fill_global_attrs(level, orbit=-1, bin_size=None, inflight=False)
       Define global attributes in the SPEXone Level-1 products.
    check_stored(allow_empty=False)
       Check variables with the same first dimension have equal sizes.
    sec_of_day(ccsds_sec, ccsds_subsec)
       Convert CCSDS timestamp to seconds after midnight.
    fill_time(ccsds_sec, ccsds_subsec, group=None)
       Write time of Science telemetry packets (UTC/TAI) to L1A product.
    fill_science(img_data, img_hk, img_id)
       Write Science data and housekeeping telemetry (Science) to L1A product.
    fill_nomhk(nomhk_data)
       Write nominal housekeeping telemetry packets (NomHK) to L1A product.
    fill_demhk(demhk_data)
       Write detector housekeeping telemetry packets (DemHK) to L1A product.
    """
    processing_level = 'L1A'
    dset_stored = {
        '/science_data/detector_images': 0,
        '/science_data/detector_telemetry': 0,
        '/image_attributes/binning_table': 0,
        '/image_attributes/digital_offset': 0,
        '/image_attributes/nr_coadditions': 0,
        '/image_attributes/exposure_time': 0,
        '/image_attributes/image_CCSDS_sec': 0,
        '/image_attributes/image_CCSDS_subsec': 0,
        '/image_attributes/image_time': 0,
        '/image_attributes/image_ID': 0,
        '/engineering_data/NomHK_telemetry': 0,
        '/engineering_data/DemHK_telemetry': 0,
        '/engineering_data/temp_detector': 0,
        '/engineering_data/temp_housing': 0,
        '/engineering_data/temp_radiator': 0,
        '/engineering_data/HK_tlm_time': 0,
        '/navigation_data/adstate': 0,
        '/navigation_data/att_quat': 0,
        '/navigation_data/orb_pos': 0,
        '/navigation_data/orb_vel': 0,
        '/navigation_data/att_time': 0,
        '/navigation_data/orb_time': 0
    }

    def close(self):
        """
        Close product and check if required datasets are filled with data
        """
        if self.fid is None:
            return

        # check if atleast one dataset is updated
        if self.fid.dimensions['number_of_images'].size == 0:
            self.fid.close()
            self.fid = None
            return

        # check of all required dataset their sizes
        self.check_stored(allow_empty=True)

        # define object to access Science telemetry parameters
        mps = TMscience(self.get_dset('/science_data/detector_telemetry')[-1])

        # determine duration master clock cycle
        imro = 1e-1 * mps.get('FTI') * 2
        mcycl = 1e-1 * mps.get('FTI') * mps.get('REG_NCOADDFRAMES')

        img_sec = self.fid['/image_attributes/image_CCSDS_sec'][:].data
        img_subsec = self.fid['/image_attributes/image_CCSDS_subsec'][:].data

        time0 = (self.epoch
                 + timedelta(seconds=int(img_sec[0]))
                 + timedelta(microseconds=int(1e6 * img_subsec[0] / 65536))
                 - timedelta(milliseconds=mcycl + imro))

        time1 = (self.epoch
                 + timedelta(seconds=int(img_sec[-1]))
                 + timedelta(microseconds=int(1e6 * img_subsec[-1] / 65536))
                 - timedelta(milliseconds=imro))

        self.fid.time_coverage_start = time0.isoformat(timespec='milliseconds')
        self.fid.time_coverage_end = time1.isoformat(timespec='milliseconds')

        self.fid.close()
        self.fid = None

    # -------------------------
    def check_stored(self, allow_empty=False):
        """
        Check variables with the same first dimension have equal sizes

        Parameters
        ----------
        allow_empty :  bool, default=False
        """
        warn_str = ('SPEX Level-1A format check [WARNING]:'
                    ' size of variable "{:s}" is wrong, only {:d} elements')

        # check image datasets
        dim_sz = self.get_dim('number_of_images')
        res = []
        key_list = [x for x in self.dset_stored
                    if (x.startswith('/science_data')
                        or x.startswith('/image_attributes'))]
        for key in key_list:
            res.append(self.dset_stored[key])
        res = np.array(res)
        if allow_empty:
            indx = ((res > 0) & (res != dim_sz)).nonzero()[0]
        else:
            indx = (res != dim_sz).nonzero()[0]
        for ii in indx:
            print(warn_str.format(key_list[ii], res[ii]))

        # check house-keeping datasets
        dim_sz = self.get_dim('hk_packets')
        key_list = [x for x in self.dset_stored
                    if x.startswith('/engineering_data')]
        res = []
        for key in key_list:
            res.append(self.dset_stored[key])
        res = np.array(res)
        if allow_empty:
            indx = ((res > 0) & (res != dim_sz)).nonzero()[0]
        else:
            indx = (res != dim_sz).nonzero()[0]
        for ii in indx:
            print(warn_str.format(key_list[ii], res[ii]))

        # check navigation datasets
        dim_sz = self.get_dim('SC_records')
        key_list = [x for x in self.dset_stored
                    if x.startswith('/navigation_data')]
        res = []
        for key in key_list:
            res.append(self.dset_stored[key])
        res = np.array(res)
        if allow_empty:
            indx = ((res > 0) & (res != dim_sz)).nonzero()[0]
        else:
            indx = (res != dim_sz).nonzero()[0]
        for ii in indx:
            print(warn_str.format(key_list[ii], res[ii]))

    # ---------- PUBLIC FUNCTIONS ----------
    def sec_of_day(self, ccsds_sec, ccsds_subsec) -> tuple:
        """
        Convert CCSDS timestamp to seconds after midnight

        Parameters
        ----------
        ccsds_sec : numpy array (dtype='u4')
          Seconds since 1970-1-1
        ccsds_subsec : numpy array (dtype='u2')
          Sub-seconds as (1 / 2**16) seconds

        Returns
        -------
        tuple holding reference_day and numpy.ndarray with sec_of_day
        """
        # determine midnight before start measurement
        tstamp0 = self.epoch + timedelta(seconds=int(ccsds_sec[0]))
        reference_day = datetime(year=tstamp0.year,
                                 month=tstamp0.month,
                                 day=tstamp0.day, tzinfo=timezone.utc)

        # store seconds since midnight
        sec_of_day = ccsds_sec - (reference_day - self.epoch).total_seconds()

        # return seconds since midnight
        return reference_day, sec_of_day + ccsds_subsec / 65536

    def fill_time(self, ccsds_sec, ccsds_subsec, group=None) -> None:
        """
        Write time of Science telemetry packets (UTC/TAI) to L1A product

        Parameters
        ----------
        ccsds_sec : numpy array (dtype='u4')
          Seconds since 1970-1-1
        ccsds_subsec : numpy array (dtype='u2')
          Sub-seconds as (1 / 2**16) seconds
        group : str, default=None

        Note
        ----
        Writes parameters: image_time, image_CCSDS_sec and image_CCSDS_subsec
        """
        if group is None:
            group = 'image_attributes'

        # calculate seconds of day
        reference_day, sec_of_day = self.sec_of_day(ccsds_sec, ccsds_subsec)

        if group in ('image_attributes', '/image_attributes'):
            self.set_dset('/image_attributes/image_CCSDS_sec', ccsds_sec)
            self.set_dset('/image_attributes/image_CCSDS_subsec', ccsds_subsec)
            self.set_dset('/image_attributes/image_time', sec_of_day)
            self.set_attr('units',
                          f'seconds since {reference_day.isoformat()}',
                          ds_name='/image_attributes/image_time')
        elif group in ('engineering_data', '/engineering_data'):
            self.set_dset('/engineering_data/HK_tlm_time', sec_of_day)
            self.set_attr('units',
                          f'seconds since {reference_day.isoformat()}',
                          ds_name='/engineering_data/HK_tlm_time')
        else:
            self.set_dset('/navigation_data/att_time', sec_of_day)
            self.set_attr('units',
                          f'seconds since {reference_day.isoformat()}',
                          ds_name='/navigation_data/att_time')
            self.set_dset('/navigation_data/orb_time', sec_of_day)
            self.set_attr('units',
                          f'seconds since {reference_day.isoformat()}',
                          ds_name='/navigation_data/orb_time')

    def fill_science(self, img_data, img_hk, img_id) -> None:
        """
        Write Science data and housekeeping telemetry (Science) to L1A product

        Parameters
        ----------
        img_data : numpy array (uint16)
           Detector image data
        img_hk : numpy array ()
           Structured array with all Science telemetry parameters
        img_id : numpy array (uint16)
           Detector image counter

        Notes
        -----
        Adds detector_telemetry data to the group /science_data

        Parameters: binning_table, digital_offset, exposure_time
        and nr_coadding are extracted from the telemetry packets and writen
        in the group /image_attributes
        """
        if len(img_hk) == 0:
            return

        self.set_dset('/science_data/detector_images', img_data)
        self.set_dset('/science_data/detector_telemetry', img_hk)

        mps = TMscience(img_hk)
        self.set_dset('/image_attributes/binning_table', mps.binning_table_id)
        self.set_dset('/image_attributes/digital_offset', mps.offset)
        self.set_dset('/image_attributes/exposure_time',
                      MCP_TO_SEC * mps.exp_time)
        self.set_dset('/image_attributes/nr_coadditions',
                      mps.get('REG_NCOADDFRAMES'))
        self.set_dset('/image_attributes/image_ID', img_id)

    def fill_nomhk(self, nomhk_data):
        """
        Write nominal house-keeping telemetry packets (NomHK) to L1A product

        Parameters
        ----------
        nomhk_data : numpy array
           Structured array with all NomHK telemetry parameters

        Notes
        -----
        Writes nomhk_data as TM_telemetry in group /engineering_data

        Parameters: temp_detector and temp_housing are extracted and converted
        to Kelvin and writen to the group /engineering_data
        """
        if len(nomhk_data) == 0:
            return

        self.set_dset('/engineering_data/NomHK_telemetry', nomhk_data)

        if np.all(nomhk_data['TS1_DEM_N_T'] == 0):
            self.set_dset('/engineering_data/temp_detector',
                          np.full(nomhk_data.size, 273))
        else:
            self.set_dset('/engineering_data/temp_detector',
                          frac_poly(nomhk_data['TS1_DEM_N_T']))

        if np.all(nomhk_data['TS2_HOUSING_N_T'] == 0):
            self.set_dset('/engineering_data/temp_housing',
                          np.full(nomhk_data.size, 293))
        else:
            self.set_dset('/engineering_data/temp_housing',
                          frac_poly(nomhk_data['TS2_HOUSING_N_T']))

        if np.all(nomhk_data['TS3_RADIATOR_N_T'] == 0):
            self.set_dset('/engineering_data/temp_radiator',
                          np.full(nomhk_data.size, 294))
        else:
            self.set_dset('/engineering_data/temp_radiator',
                          frac_poly(nomhk_data['TS3_RADIATOR_N_T']))

    def fill_demhk(self, demhk_data):
        """
        Write detector housekeeping telemetry packets (DemHK) to L1A product

        Parameters
        ----------
        demhk_data : numpy array
           Structured array with all DemHK telemetry parameters

        Notes
        -----
        Writes demhk_data as DetTM_telemetry in group /engineering_data

        Parameters: temp_detector and temp_housing are extracted and converted
        to Kelvin and writen to the group /engineering_data
        """
        if len(demhk_data) == 0:
            return

        self.set_dset('/engineering_data/DemHK_telemetry', demhk_data)


# - class L1Bio -------------------------
class L1Bio(Lv1io):
    """
    This class can be used to create a SPEXone Level-1B product

    Parameters
    ----------
    lv1_product: string
       Name of the Level-1B product
    append : boolean, default=False
       If the file is opened in append mode, then parameter 'dims' is ignored
    number_of_images: int, default=None
       Number of images used as input to generate the L1B product.
       This dimension is by default UNLIMITED.
    spatial_samples: int, default=200
       Total number of spatial samples from all viewport

    Attributes
    ----------
    product: pathlib.Path object
       Concrete path object to SPEXone Level-1 product
    fid: netCDF5.Dataset object
       NetCDF4 Pointer to SPEXone Level-1 product
    dset_stored: dict
       Number of items stored for all required netCDF4 variables

    Methods
    -------
    close()
       Close product and check if required datasets are filled with data.
    epoch
       Provide epoch for SPEXone.
    get_dim(ds_name)
       Returns size of dimension.
    get_attr(attr_name, ds_name=None)
       Read data of an attribute, global or attached to a group or variable.
    set_attr(attr_name, value, ds_name=None)
       Write data to an attribute, global or attached to a group or variable.
    get_dset(ds_name)
       Read data of a netCDF4 variable.
    set_dset(ds_name, value, ibgn=-1)
       Write/append data to a netCDF4 variable.
    fill_global_attrs(level, orbit=-1, bin_size=None)
       Define global attributes in the SPEXone Level-1 products.
    check_stored()
       Check variables with the same first dimension have equal sizes.

    Notes
    -----
    ToDo: make sure we store the reference date for image_time
    """
    processing_level = 'L1B'
    dset_stored = {
        '/BIN_ATTRIBUTES/image_time': 0,
        '/GEOLOCATION_DATA/altitude': 0,
        '/GEOLOCATION_DATA/latitude': 0,
        '/GEOLOCATION_DATA/longitude': 0,
        '/GEOLOCATION_DATA/sensor_azimuth': 0,
        '/GEOLOCATION_DATA/sensor_zenith': 0,
        '/GEOLOCATION_DATA/solar_azimuth': 0,
        '/GEOLOCATION_DATA/solar_zenith': 0,
        '/OBSERVATION_DATA/I': 0,
        '/OBSERVATION_DATA/I_noise': 0,
        '/OBSERVATION_DATA/AoLP': 0,
        '/OBSERVATION_DATA/AoLP_noise': 0,
        '/OBSERVATION_DATA/DoLP': 0,
        '/OBSERVATION_DATA/DoLP_noise': 0,
        '/OBSERVATION_DATA/Q_over_I': 0,
        '/OBSERVATION_DATA/Q_over_I_noise': 0,
        '/OBSERVATION_DATA/U_over_I': 0,
        '/OBSERVATION_DATA/U_over_I_noise': 0,
        '/SENSOR_VIEWS_BANDS/viewport_index': 0,
        '/SENSOR_VIEWS_BANDS/intensity_wavelengths': 0,
        '/SENSOR_VIEWS_BANDS/intensity_bandpasses': 0,
        '/SENSOR_VIEWS_BANDS/intensity_F0': 0,
        '/SENSOR_VIEWS_BANDS/polarization_wavelengths': 0,
        '/SENSOR_VIEWS_BANDS/polarization_bandpasses': 0,
        '/SENSOR_VIEWS_BANDS/polarization_F0': 0,
        '/SENSOR_VIEWS_BANDS/view_angles': 0
    }

    def close(self):
        """
        Close product and check if required datasets are filled with data
        """
        if self.fid is None:
            return

        # check if atleast one dataset is updated
        if self.fid.dimensions['bins_along_track'].size == 0:
            self.fid.close()
            self.fid = None
            return

        # check of all required dataset their sizes
        self.check_stored()

        # update coverage time
        secnd = self.fid['/BIN_ATTRIBUTES/image_time'][0].data
        time0 = (self.epoch
                 + timedelta(seconds=int(secnd))
                 + timedelta(microseconds=int(secnd % 1)))

        secnd = self.fid['/BIN_ATTRIBUTES/image_time'][-1].data
        time1 = (self.epoch
                 + timedelta(seconds=int(secnd))
                 + timedelta(microseconds=int(secnd % 1)))

        self.fid.time_coverage_start = time0.isoformat(timespec='milliseconds')
        self.fid.time_coverage_end = time1.isoformat(timespec='milliseconds')

        self.fid.close()
        self.fid = None

    # -------------------------
    def check_stored(self):
        """
        Check variables with the same first dimension have equal sizes
        """
        warn_str = ('SPEX Level-1B format check [WARNING]:'
                    ' size of variable "{:s}" is wrong, only {:d} elements')

        # check datasets in group /SENSOR_VIEWS_BANDS
        dim_sz = self.get_dim('number_of_views')
        res = []
        key_list = [x for x in self.dset_stored
                    if x.startswith('/SENSOR_VIEWS_BANDS')]
        for key in key_list:
            if key == '/SENSOR_VIEWS_BANDS/viewport_index':
                continue
            res.append(self.dset_stored[key])
        res = np.array(res)
        indx = (res != dim_sz).nonzero()[0]
        for ii in indx:
            print(warn_str.format(key_list[ii], res[ii]))

        # check datasets in all other groups
        dim_sz = self.get_dim('bins_along_track')
        res = []
        key_list = [x for x in self.dset_stored
                    if not x.startswith('/SENSOR_VIEWS_BANDS')]
        for key in key_list:
            res.append(self.dset_stored[key])
        res = np.array(res)
        indx = (res != dim_sz).nonzero()[0]
        for ii in indx:
            print(warn_str.format(key_list[ii], res[ii]))

        for ii, key in enumerate(self.dset_stored):
            print(ii, key, self.dset_stored[key])

    # ---------- PUBLIC FUNCTIONS ----------


# - class L1Cio -------------------------
class L1Cio(Lv1io):
    """
    This class can be used to create a SPEXone Level-1C product

    Parameters
    ----------
    lv1_product :  str
       Name of the Level-1C product
    append :  bool, default=False
       If the file is opened in append mode, then parameter 'dims' is ignored
    number_of_images: int, default=None
       Number of images used as input to generate the L1B product.
       This dimension is by default UNLIMITED.

    Attributes
    ----------
    product: pathlib.Path object
       Concrete path object to SPEXone Level-1 product
    inflight: boolean
       Flag to indicate data collected during in-flight of on-ground
    fid: netCDF5.Dataset object
       NetCDF4 Pointer to SPEXone Level-1 product
    dset_stored: dict
       Number of items stored for all required netCDF4 variables

    Methods
    -------
    close()
       Close product and check if required datasets are filled with data.
    epoch
       Provide epoch for SPEXone.
    get_dim(ds_name)
       Returns size of dimension.
    get_attr(attr_name, ds_name=None)
       Read data of an attribute, global or attached to a group or variable.
    set_attr(attr_name, value, ds_name=None)
       Write data to an attribute, global or attached to a group or variable.
    get_dset(ds_name)
       Read data of a netCDF4 variable.
    set_dset(ds_name, value)
       Write/append data to a netCDF4 variable.
    fill_global_attrs(level, orbit=-1, bin_size=None)
       Define global attributes in the SPEXone Level-1 products.
    check_stored()
       Check variables with the same first dimension have equal sizes.

    Notes
    -----
    ToDo: make sure we store the reference date for image_time
    """
    processing_level = 'L1C'
    dset_stored = {
        '/BIN_ATTRIBUTES/nadir_view_time': 0,
        '/BIN_ATTRIBUTES/view_time_offsets': 0,
        '/GEOLOCATION_DATA/latitude': 0,
        '/GEOLOCATION_DATA/longitude': 0,
        '/GEOLOCATION_DATA/altitude': 0,
        '/GEOLOCATION_DATA/altitude_variability': 0,
        '/GEOLOCATION_DATA/sensor_azimuth': 0,
        '/GEOLOCATION_DATA/sensor_zenith': 0,
        '/GEOLOCATION_DATA/solar_azimuth': 0,
        '/GEOLOCATION_DATA/solar_zenith': 0,
        '/OBSERVATION_DATA/obs_per_view': 0,
        '/OBSERVATION_DATA/AoLP': 0,
        '/OBSERVATION_DATA/AoLP_noise': 0,
        '/OBSERVATION_DATA/DoLP': 0,
        '/OBSERVATION_DATA/DoLP_noise': 0,
        '/OBSERVATION_DATA/I': 0,
        '/OBSERVATION_DATA/I_noise': 0,
        '/OBSERVATION_DATA/I_polsample': 0,
        '/OBSERVATION_DATA/I_polsample_noise': 0,
        '/OBSERVATION_DATA/QC': 0,
        '/OBSERVATION_DATA/QC_bitwise': 0,
        '/OBSERVATION_DATA/QC_polsample': 0,
        '/OBSERVATION_DATA/QC_polsample_bitwise': 0,
        '/OBSERVATION_DATA/Q_over_I': 0,
        '/OBSERVATION_DATA/Q_over_I_noise': 0,
        '/OBSERVATION_DATA/U_over_I': 0,
        '/OBSERVATION_DATA/U_over_I_noise': 0,
        '/SENSOR_VIEWS_BANDS/intensity_bandpasses': 0,
        '/SENSOR_VIEWS_BANDS/intensity_wavelengths': 0,
        '/SENSOR_VIEWS_BANDS/intensity_F0': 0,
        '/SENSOR_VIEWS_BANDS/polarization_bandpasses': 0,
        '/SENSOR_VIEWS_BANDS/polarization_wavelengths': 0,
        '/SENSOR_VIEWS_BANDS/polarization_F0': 0,
        '/SENSOR_VIEWS_BANDS/view_angles': 0
    }

    def close(self):
        """
        Close product and check if required datasets are filled with data
        """
        if self.fid is None:
            return

        # check if atleast one dataset is updated
        if self.fid.dimensions['bins_along_track'].size == 0:
            self.fid.close()
            self.fid = None
            return

        # check of all required dataset their sizes
        self.check_stored()

        # update coverage time
        secnd = self.fid['/BIN_ATTRIBUTES/nadir_view_time'][0].data
        time0 = (self.epoch
                 + timedelta(seconds=int(secnd))
                 + timedelta(microseconds=int(secnd % 1)))

        secnd = self.fid['/BIN_ATTRIBUTES/nadir_view_time'][-1].data
        time1 = (self.epoch
                 + timedelta(seconds=int(secnd))
                 + timedelta(microseconds=int(secnd % 1)))

        self.fid.time_coverage_start = time0.isoformat(timespec='milliseconds')
        self.fid.time_coverage_end = time1.isoformat(timespec='milliseconds')

        self.fid.close()
        self.fid = None

    # -------------------------
    def check_stored(self):
        """
        Check variables with the same first dimension have equal sizes
        """
        warn_str = ('SPEX Level-1C format check [WARNING]:'
                    ' size of variable "{:s}" is wrong, only {:d} elements')

        # check datasets in group /SENSOR_VIEWS_BANDS
        dim_sz = self.get_dim('number_of_views')
        res = []
        key_list = [x for x in self.dset_stored
                    if x.startswith('/SENSOR_VIEWS_BANDS')]
        for key in key_list:
            if key == '/SENSOR_VIEWS_BANDS/viewport_index':
                continue
            res.append(self.dset_stored[key])
        res = np.array(res)
        indx = (res != dim_sz).nonzero()[0]
        for ii in indx:
            print(warn_str.format(key_list[ii], res[ii]))

        # check datasets in all other groups
        dim_sz = self.get_dim('bins_along_track')
        res = []
        key_list = [x for x in self.dset_stored
                    if not x.startswith('/SENSOR_VIEWS_BANDS')]
        for key in key_list:
            res.append(self.dset_stored[key])
        res = np.array(res)
        indx = (res != dim_sz).nonzero()[0]
        for ii in indx:
            print(warn_str.format(key_list[ii], res[ii]))

        for ii, key in enumerate(self.dset_stored):
            print(ii, key, self.dset_stored[key])

    # ---------- PUBLIC FUNCTIONS ----------



"""
This file is part of pyspex

https://github.com/rmvanhees/pyspex.git

Determine if measurements are performed inflight or on-ground.

Copyright (c) 2021 SRON - Netherlands Institute for Space Research
   All Rights Reserved

License:  BSD-3-Clause
"""
from datetime import datetime, timezone

# The on-ground calibration campaigns are expected to be finished long
# before the expected launch date of PACE in 2023
LAUNCH_DATE = datetime(2023, 1, 1, tzinfo=timezone.utc)


def before_launch(timestamp: int) -> bool:
    """
    Return True when timestamp is before launch of PACE
    """
    return timestamp < LAUNCH_DATE



#!/usr/bin/env python3
"""
This file is part of pyspex

https://github.com/rmvanhees/pyspex.git

Create new SPEXone Level-1B product with selected data from original

Copyright (c) 2021 SRON - Netherlands Institute for Space Research
   All Rights Reserved

License:  BSD-3-Clause
"""
import argparse
from datetime import datetime
from pathlib import Path

import h5py

from pyspex.lv1_io import L1Bio


# --------------------------------------------------
def main():
    """
    Main function of this module
    """
    parser = argparse.ArgumentParser(
        description=('Copy selected data from one SPEXone L1B product'
                     ' into a new SPEXone L1B product'))
    parser.add_argument('--verbose', '-v', action='store_true',
                        help='be verbose, default be silent')
    parser.add_argument('--mps_id', nargs='*', type=int, default=None,
                        help='select on MPS-ID')
    # parser.add_argument('--time', nargs=2, default=None,
    #                    help='select on image time [start, end]')
    # parser.add_argument('--', default=None, help='')
    parser.add_argument('--out', default='.',
                        help=('name of directory to store the new Level-1B'
                              ' product, default: current working directory'))
    parser.add_argument('l1b_product', default=None,
                        help='name of SPEXone Level-1B product')
    args = parser.parse_args()
    if args.verbose:
        print(args)

    l1b_product = Path(args.l1b_product)
    if not l1b_product.is_file():
        raise FileNotFoundError(f'File {args.l1b_product} does not exist')
    # ToDo: check if SPEXone Level-1B product
    # ToDo: implement check on data product

    out_dir = Path(args.out_dir)
    if not out_dir.is_dir():
        out_dir.mkdir(mode=0o755, parents=True)

    # ----- read data from orignal product -----
    # pylint: disable=no-member, unsubscriptable-object
    with h5py.File(l1b_product) as fid:
        # group BIN_ATTRIBUTES
        ref_date = None
        image_time = fid['/BIN_ATTRIBUTES/image_time'][:]
        units = fid['/BIN_ATTRIBUTES/image_time'].attrs['units']
        if len(units) > 14:
            ref_date = datetime.fromisoformat(units[14:])
        elif fid.attrs['time_coverage_start'] != b'Unknown':
            ref_date = datetime.fromisoformat(
                fid.attrs['time_coverage_start'].decode('ascii'))
            ref_date = ref_date.replace(hour=0, minute=0, second=0)

        # group GEOLOCATION_DATA
        altitude = fid['/GEOLOCATION_DATA/altitude'][:]
        latitude = fid['/GEOLOCATION_DATA/latitude'][:]
        longitude = fid['/GEOLOCATION_DATA/longitude'][:]
        sensor_azi = fid['/GEOLOCATION_DATA/sensor_azimuth'][:]
        sensor_zen = fid['/GEOLOCATION_DATA/sensor_zenith'][:]
        solar_azi = fid['/GEOLOCATION_DATA/solar_azimuth'][:]
        solar_zen = fid['/GEOLOCATION_DATA/solar_zenith'][:]

        # group OBSERVATION_DATA
        intens_val = fid['/OBSERVATION_DATA/I'][:]
        intens_noise = fid['/OBSERVATION_DATA/I_noise'][:]
        aolp_val = fid['/OBSERVATION_DATA/AoLP'][:]
        aolp_noise = fid['/OBSERVATION_DATA/AoLP_noise'][:]
        dolp_val = fid['/OBSERVATION_DATA/DoLP'][:]
        dolp_noise = fid['/OBSERVATION_DATA/DoLP_noise'][:]
        q_val = fid['/OBSERVATION_DATA/q'][:]
        q_noise = fid['/OBSERVATION_DATA/q_noise'][:]
        u_val = fid['/OBSERVATION_DATA/u'][:]
        u_noise = fid['/OBSERVATION_DATA/u_noise'][:]

        # group SENSOR_VIEWS_BANDS
        group = 'SENSOR_VIEWS_BANDS' if 'SENSOR_VIEWS_BANDS' in fid \
            else 'SENSOR_VIEW_BANDS'
        view_indx = fid[f'/{group}/viewport_index'][:]
        intens_bands = fid[f'/{group}/intensity_bandpasses'][:]
        intens_wav = fid[f'/{group}/intensity_wavelengths'][:]
        intens_f0 = fid[f'/{group}/intensity_f0'][:]
        polar_bands = fid[f'/{group}/polarization_bandpasses'][:]
        polar_wav = fid[f'/{group}/polarization_wavelengths'][:]
        polar_f0 = fid[f'/{group}/polarization_f0'][:]
        view_angles = fid[f'/{group}/view_angles'][:]

        # global attributes
        global_attrs = {}
        check_attrs = ("history", "sun_earth_distance",
                       "terrain_data_source",
                       "spectral_response_function",
                       "systematic_uncertainty_model")
        for key in sorted(check_attrs):
            if key in fid.attrs:
                if isinstance(fid.attrs[key], h5py.Empty):
                    global_attrs[key] = ""
                else:
                    global_attrs[key] = fid.attrs[key].decode('ascii')

    dims = {'bins_along_track':  latitude.shape[0],
            'spatial_samples_per_image':  latitude.shape[1],
            'intensity_bands_per_view':  intens_bands.shape[1],
            'polarization_bands_per_view':  polar_bands.shape[1]}

    # ----- perform data selection -----
    # ToDo: implement data selection

    # ----- now we can update the name of the output product -----
    # - because the production time has changed
    # - and when coverage time is changed
    if (out_dir / l1b_product.name).is_file() \
       and l1b_product.samefile(out_dir / l1b_product.name):
        raise OSError('Output will overwrite original product')

    # ----- write new output product with selected data -----
    print('dims: ', dims)
    print('ref_date: ', ref_date)
    with L1Bio(out_dir / l1b_product.name, ref_date=ref_date, dims=dims) as l1b:
        # group BIN_ATTRIBUTES
        l1b.set_dset('/BIN_ATTRIBUTES/image_time', image_time)

        # group GEOLOCATION_DATA
        l1b.set_dset('/GEOLOCATION_DATA/altitude', altitude)
        l1b.set_dset('/GEOLOCATION_DATA/latitude', latitude)
        l1b.set_dset('/GEOLOCATION_DATA/longitude', longitude)
        l1b.set_dset('/GEOLOCATION_DATA/sensor_azimuth', sensor_azi)
        l1b.set_dset('/GEOLOCATION_DATA/sensor_zenith', sensor_zen)
        l1b.set_dset('/GEOLOCATION_DATA/solar_azimuth', solar_azi)
        l1b.set_dset('/GEOLOCATION_DATA/solar_zenith', solar_zen)

        # group OBSERVATION_DATA
        l1b.set_dset('/OBSERVATION_DATA/I', intens_val)
        l1b.set_dset('/OBSERVATION_DATA/I_noise', intens_noise)
        l1b.set_dset('/OBSERVATION_DATA/AoLP', aolp_val)
        l1b.set_dset('/OBSERVATION_DATA/AoLP_noise', aolp_noise)
        l1b.set_dset('/OBSERVATION_DATA/DoLP', dolp_val)
        l1b.set_dset('/OBSERVATION_DATA/DoLP_noise', dolp_noise)
        l1b.set_dset('/OBSERVATION_DATA/Q_over_I', q_val)
        l1b.set_dset('/OBSERVATION_DATA/Q_over_I_noise', q_noise)
        l1b.set_dset('/OBSERVATION_DATA/U_over_I', u_val)
        l1b.set_dset('/OBSERVATION_DATA/U_over_I_noise', u_noise)

        # group SENSOR_VIEWS_BANDS
        l1b.set_dset('/SENSOR_VIEWS_BANDS/viewport_index', view_indx)
        l1b.set_dset('/SENSOR_VIEWS_BANDS/intensity_bandpasses', intens_bands)
        l1b.set_dset('/SENSOR_VIEWS_BANDS/intensity_wavelengths', intens_wav)
        l1b.set_dset('/SENSOR_VIEWS_BANDS/intensity_F0', intens_f0)
        l1b.set_dset('/SENSOR_VIEWS_BANDS/polarization_bandpasses', polar_bands)
        l1b.set_dset('/SENSOR_VIEWS_BANDS/polarization_wavelengths', polar_wav)
        l1b.set_dset('/SENSOR_VIEWS_BANDS/polarization_F0', polar_f0)
        l1b.set_dset('/SENSOR_VIEWS_BANDS/view_angles', view_angles)

        # write global attributes
        l1b.fill_global_attrs(inflight=True)
        for key, value in global_attrs.items():
            l1b.set_attr(key, value)


# --------------------------------------------------
if __name__ == '__main__':
    main()



"""
This file is part of pyspex

https://github.com/rmvanhees/pyspex.git

Defines global attributes for SPEXone L1 products

Copyright (c) 2020-2021 SRON - Netherlands Institute for Space Research
   All Rights Reserved

License:  BSD-3-Clause
"""
from datetime import datetime, timezone

from pyspex import version


# - local functions --------------------------------
def attrs_def(level: str, inflight=True, origin=None) -> dict:
    """
    Defines all global attributes for SPEXone Level-1 products

    Parameters
    ----------
    level : str
      Product processing level 'L1A', 'L1B' or 'L1C'
    inflight : bool
      Flag for in-flight or on-ground products
    origin : str
      Product origin: 'SRON' or 'NASA'

    Returns
    -------
    dict
    """
    if origin is None:
        origin = 'NASA' if inflight else 'SRON'

    res = {
        "title": f"PACE SPEX Level-{level[1:]} data",
        "instrument": "SPEX",
        "processing_version": "V1.0",
        "conventions": "CF-1.6",
        "institution": ("NASA Goddard Space Flight Center,"
                        " Ocean Biology Processing Group"),
        "license": ("http://science.nasa.gov/earth-science/"
                    "earth-science-data/data-information-policy/"),
        "naming_authority": "gov.nasa.gsfc.sci.oceancolor",
        "keyword_vocabulary": ("NASA Global Change Master Directory (GCMD)"
                               " Science Keywords"),
        "stdname_vocabulary": ("NetCDF Climate and Forecast (CF)"
                               " Metadata Convention"),
        "creator_name": "NASA/GSFC",
        "creator_email": "data@oceancolor.gsfc.nasa.gov",
        "creator_url": "http://oceancolor.gsfc.nasa.gov",
        "project": "PACE Project",
        "publisher_name": "NASA/GSFC",
        "publisher_email": "data@oceancolor.gsfc.nasa.gov",
        "publisher_url": "http://oceancolor.gsfc.nasa.gov",
        "processing_level": level,
        "CDM_data_type": "swath" if inflight else "on-ground calibration",
        "orbit_number": -1,
        "history": None,
        "CDL_version_date": "2020-02-20",
        "product_name": None,
        "start_direction": "Ascending" if inflight else None,
        "end_direction": "Ascending" if inflight else None,
        "time_coverage_start": "yyyy-mm-ddTHH:MM:DD",
        "time_coverage_end": "yyyy-mm-ddTHH:MM:DD",
        "date_created": datetime.now(timezone.utc).isoformat(
            timespec='milliseconds'),
        "software_name": 'https://github.com/rmvanhees/pyspex',
        "software_version": version.get(),
        "sun_earth_distance": None,
        "terrain_data_source": None,
        "spectral_response_function": None,
        "systematic_uncertainty_model": None,
        "nadir_bin": None,
        "bin_size_at_nadir": None
    }

    if origin == 'SRON':
        res['title'] = f"SPEXone Level-{level[1:]} data"
        res['instrument'] = "SPEXone"
        res['institution'] = "SRON Netherlands Institute for Space Research"
        res['creator_name'] = "SRON/Earth Science"
        res['creator_email'] = "SPEXone-MPC@sron.nl"
        res['creator_url'] = "https://www.sron.nl/missions-earth/pace-spexone"
        res['publisher_name'] = "SRON/Earth Science"
        res['publisher_email'] = "SPEXone-MPC@sron.nl"
        res['publisher_url'] = "https://www.sron.nl/missions-earth/pace-spexone"

    return res



"""
This file is part of pyspex

https://github.com/rmvanhees/pyspex.git

Copyright (c) 2021 SRON - Netherlands Institute for Space Research
   All Rights Reserved

License:  BSD-3-Clause
"""
from pyspex import __version__


def get(full=False):
    """
    Returns software version as obtained from git
    """
    if full:
        return __version__

    return __version__.split('+')[0]



"""
This file is part of pyspex

https://github.com/rmvanhees/pyspex.git

Python implementation of the PACE SPEX Level-1A product (inflight and onground)

Copyright (c) 2020-2021 SRON - Netherlands Institute for Space Research
   All Rights Reserved

License:  BSD-3-Clause
"""
import numpy as np

from netCDF4 import Dataset

from .tmtc_def import tmtc_def

# - global parameters ------------------------------


# - local functions --------------------------------


# - main function ----------------------------------
def init_l1a(l1a_flname: str, dims: dict) -> None:
    """
    Create an empty SPEXone Level-1A product (on-ground or in-flight)

    Parameters
    ----------
    l1a_flname : string
       Name of L1A product
    dims :   dictionary
       Provide length of the Level-1A dimensions
       Default values:
            number_of_images : None     # number of image frames
            samples_per_image : 184000  # depends on binning table
            SC_records : None           # space-craft navigation records (1 Hz)
            hk_packets : None           # number of HK tlm-packets (1 Hz)

    Notes
    -----
    Original design Frederick S. Patt (Goddard Space Flight Center), 08-Feb-2019
    Modified for on-ground calibration measurements, Richard M. van Hees (SRON)
    """
    # check function parameters
    if not isinstance(dims, dict):
        raise TypeError("dims should be a dictionary")

    # initialize dimensions
    number_img = None
    img_samples = 184000
    hk_packets = None
    sc_records = None

    if 'number_of_images' in dims:
        number_img = dims['number_of_images']
    if 'samples_per_image' in dims:
        img_samples = dims['samples_per_image']
    if 'SC_records' in dims:
        sc_records = dims['SC_records']
    if 'hk_packets' in dims:
        hk_packets = dims['hk_packets']

    # create/overwrite netCDF4 product
    rootgrp = Dataset(l1a_flname, 'w')

    # - define global dimensions
    _ = rootgrp.createDimension('number_of_images', number_img)
    _ = rootgrp.createDimension('samples_per_image', img_samples)
    _ = rootgrp.createDimension('hk_packets', hk_packets)
    _ = rootgrp.createDimension('SC_records', sc_records)
    _ = rootgrp.createDimension('quaternion_elements', 4)
    _ = rootgrp.createDimension('vector_elements', 3)

    # - define group /image_attributs and its datasets
    sgrp = rootgrp.createGroup('/image_attributes')
    dset = sgrp.createVariable('image_time', 'f8', ('number_of_images',))
    dset.long_name = "image time (seconds of day)"
    dset.valid_min = 0
    dset.valid_max = 86400.999999
    dset.units = "seconds since midnight"
    dset = sgrp.createVariable('image_CCSDS_sec', 'u4',
                               ('number_of_images',))
    dset.long_name = "image CCSDS time (seconds)"
    dset.valid_min = np.uint32(1577500000)  # year 2020
    dset.valid_max = np.uint32(2050000000)  # year 2035
    dset.units = "seconds since 1970-1-1 0:0:0"             # UTC or TAI?
    dset = sgrp.createVariable('image_CCSDS_subsec', 'u2',
                               ('number_of_images',))
    dset.long_name = "image CCSDS time (sub-seconds)"
    dset.valid_min = np.uint16(0)
    dset.valid_max = np.uint16(0xFFFF)
    dset.units = "1 / 65536 seconds"
    dset = sgrp.createVariable('image_ID', 'i4', ('number_of_images',))
    dset.long_name = "image counter from power-up"
    dset.valid_min = np.int32(0)
    dset.valid_max = np.int32(0x7FFFFFFF)
    dset = sgrp.createVariable('binning_table', 'u1', ('number_of_images',))
    dset.long_name = "binning-table ID"
    dset.valid_min = np.uint8(0)
    dset.valid_max = np.uint8(0xFF)
    dset = sgrp.createVariable('digital_offset', 'i2', ('number_of_images',))
    dset.long_name = "digital offset"
    dset.units = "1"
    dset = sgrp.createVariable('nr_coadditions', 'u2', ('number_of_images',),
                               fill_value=0)
    dset.long_name = "number of coadditions"
    dset.valid_min = np.int32(1)
    dset.units = "1"
    dset = sgrp.createVariable('exposure_time', 'f8', ('number_of_images',),
                               fill_value=0)
    dset.long_name = "exposure time"
    dset.units = "sec"

    # - define group /science_data and its datasets
    sgrp = rootgrp.createGroup('/science_data')
    chunksizes = None if number_img is not None else (1, img_samples)
    dset = sgrp.createVariable('detector_images', 'u2',
                               ('number_of_images', 'samples_per_image'),
                               chunksizes=chunksizes, fill_value=0xFFFF)
    dset.long_name = "Image data from detector"
    dset.valid_min = np.uint16(0)
    dset.valid_max = np.uint16(0xFFFE)
    dset.units = "counts"
    hk_dtype = rootgrp.createCompoundType(np.dtype(tmtc_def(0x350)),
                                          'science_dtype')
    dset = sgrp.createVariable('detector_telemetry', hk_dtype,
                               dimensions=('number_of_images',))
    dset.long_name = "SPEX science telemetry"
    dset.comment = "a subset of MPS and housekeeping parameters"

    # - define group /engineering_data and its datasets
    sgrp = rootgrp.createGroup('/engineering_data')
    dset = sgrp.createVariable('HK_tlm_time', 'f8', ('hk_packets',))
    dset.long_name = "HK telemetry packet time (seconds of day)"
    dset.valid_min = 0
    dset.valid_max = 86400.999999
    dset.units = "seconds since midnight"
    hk_dtype = rootgrp.createCompoundType(np.dtype(tmtc_def(0x320)),
                                          'nomhk_dtype')
    dset = sgrp.createVariable('NomHK_telemetry', hk_dtype, ('hk_packets',))
    dset.long_name = "SPEX nominal-HK telemetry"
    dset.comment = "an extended subset of the housekeeping parameters"
    dset = sgrp.createVariable('temp_detector', 'f4', ('hk_packets',))
    dset.long_name = "Detector temperature"
    dset.comment = "TS1 DEM Temperature (nominal)"
    dset.valid_min = 260
    dset.valid_max = 300
    dset.units = "K"
    dset = sgrp.createVariable('temp_housing', 'f4', ('hk_packets',))
    dset.long_name = "Housing temperature"
    dset.comment = "TS2 Housing Temperature (nominal)"
    dset.valid_min = 260
    dset.valid_max = 300
    dset.units = "K"
    dset = sgrp.createVariable('temp_radiator', 'f4', ('hk_packets',))
    dset.long_name = "Radiator temperature"
    dset.comment = "TS3 Radiator Temperature (nominal)"
    dset.valid_min = 260
    dset.valid_max = 300
    dset.units = "K"
    hk_dtype = rootgrp.createCompoundType(np.dtype(tmtc_def(0x322)),
                                          'demhk_dtype')
    dset = sgrp.createVariable('DemHK_telemetry', hk_dtype, ('hk_packets',))
    dset.long_name = "SPEX detector-HK telemetry"
    dset.comment = "DEM housekeeping parameters"

    # - define group /navigation_data and its datasets
    sgrp = rootgrp.createGroup('/navigation_data')
    dset = sgrp.createVariable('adstate', 'u1',
                               ('SC_records',), fill_value=0xFF)
    dset.long_name = "Current ADCS State"
    dset.flag_values = np.array([0, 1, 2, 3, 4, 5], dtype='u1')
    dset.flag_meanings = "Wait Detumple AcqSun Point Delta Earth"
    dset = sgrp.createVariable('att_time', 'f8', ('SC_records',))
    dset.long_name = "Attitude sample time (seconds of day)"
    dset.valid_min = 0.0
    dset.valid_max = 86400.999999
    dset.units = "seconds since midnight"
    chunksizes = None if sc_records is not None else (256, 4)
    dset = sgrp.createVariable('att_quat', 'f4',
                               ('SC_records', 'quaternion_elements'),
                               chunksizes=chunksizes)
    dset.long_name = "Attitude quaternions (J2000 to spacecraft)"
    dset.valid_min = -1
    dset.valid_max = 1
    dset.units = "1"
    dset = sgrp.createVariable('orb_time', 'f8', ('SC_records',))
    dset.long_name = "Orbit vector time (seconds of day)"
    dset.valid_min = 0
    dset.valid_max = 86400.999999
    dset.units = "seconds since midnight"
    chunksizes = None if sc_records is not None else (340, 3)
    dset = sgrp.createVariable('orb_pos', 'f4',
                               ('SC_records', 'vector_elements'),
                               chunksizes=chunksizes)
    dset.long_name = "Orbit positions vectors (J2000)"
    dset.valid_min = -7200000
    dset.valid_max = 7200000
    dset.units = "meters"
    dset = sgrp.createVariable('orb_vel', 'f4',
                               ('SC_records', 'vector_elements'),
                               chunksizes=chunksizes)
    dset.long_name = "Orbit velocity vectors (J2000)"
    dset.valid_min = -7600
    dset.valid_max = 7600
    dset.units = "meters s-1"

    return rootgrp



# This file is part of pyspex
#
# https://github.com/rmvanhees/pyspex.git
#
# Copyright (c) 2019-2021 SRON - Netherlands Institute for Space Research
#   All Rights Reserved
#
# License:  BSD-3-Clause

"""Python package pyspex contains software to access
   and create SPEXone L1A and L1B products.
"""

from pkg_resources import get_distribution, DistributionNotFound

try:
    __version__ = get_distribution(__name__).version
except DistributionNotFound:
    pass



#!/usr/bin/env python3
"""
This file is part of pyspex

https://github.com/rmvanhees/pyspex.git

Python implementation to convert SPEXone DEM measurements to L1A format

Copyright (c) 2019-2021 SRON - Netherlands Institute for Space Research
   All Rights Reserved

License:  BSD-3-Clause
"""
import argparse
from datetime import datetime, timezone
from pathlib import Path

import h5py
import numpy as np

from pyspex import spx_product
from pyspex.lib.tmtc_def import tmtc_def
from pyspex.dem_io import DEMio
from pyspex.lv1_io import L1Aio
from pyspex.lv1_gse import LV1gse

# - global parameters ------------------------------
EPOCH = datetime(1970, 1, 1, tzinfo=timezone.utc)


# - local functions --------------------------------


# - main function ----------------------------------
def main():
    """
    main program to illustate the creation of a L1A calibration product
    """
    parser = argparse.ArgumentParser(
        description=('create SPEXone L1A product from DEM measurement(s)'))
    parser.add_argument('--verbose', default=False, action='store_true',
                        help='be verbose, default be silent')
    parser.add_argument('--reference', default=None,
                        help='reference detector data (for non-linearity)')
    parser.add_argument('--output', default=None,
                        help='define output directory, default=CWD')
    parser.add_argument('--dem_id', choices=('D35', 'D39'), default=None,
                        help=('provide DEM ID'
                              ' or ID will be extracted from path'))
    parser.add_argument('file_list', nargs='+',
                        help=("provide path to _b.bin file of one or more"
                              " SPEXone DEM characterization measurements"))
    args = parser.parse_args()
    if args.verbose:
        print(args)

    # sort file on timestamp in filename (problem: midnight...)
    buff = [name for name in args.file_list if name.endswith('.bin')]
    args.file_list = sorted(buff, key=lambda name: name.split('_')[-2])

    # obtain DEM_ID (can be specified on command-line)
    # and list of measurements used to generate the L1A product
    list_dem_id = []
    for file in args.file_list:
        parts = Path(file).parts
        if len(parts) > 2:
            list_dem_id.append(parts[-2].split('_')[1])

    if args.dem_id is None:
        dem_id_list = []
        for file in args.file_list:
            parts = Path(file).parts
            if len(parts) == 1:
                continue
            for dem_id in ('D35', 'D39'):
                if parts[-2].find(dem_id) != -1:
                    dem_id_list.append(dem_id)

        if not dem_id_list:
            raise KeyError(
                'Can not determine DEM_ID, please specify --dem_id')
        dem_id_list = set(dem_id_list)
        if len(dem_id_list) != 1:
            raise KeyError('DEM_ID of measurements must be unique')

        args.dem_id = dem_id_list.pop()

    # initialze data arrays
    msm_id = None
    tstamp = []
    image_list = []

    # Measurement Parameters Settings
    n_images = len(args.file_list)
    img_hk = np.zeros(n_images, dtype=np.dtype(tmtc_def(0x350)))
    hk_data = np.zeros(n_images, dtype=np.dtype(tmtc_def(0x320)))
    t_exp = np.empty(n_images, dtype=float)
    t_frm = np.empty(n_images, dtype=float)
    offset = np.empty(n_images, dtype=float)
    temp = np.empty(n_images, dtype=float)

    for ii, flname in enumerate(args.file_list):
        dem_file = Path(flname)
        if not dem_file.is_file():
            raise FileNotFoundError(f'file {flname} does not exist')
        if dem_file.suffix == '.txt':
            raise ValueError("We should not try a file with suffix '.txt'")

        # read DEM settings and data
        parts = dem_file.name.split('_')
        if len(parts) == 8:
            _id = '_'.join(parts[:4])
            tstamp.append(datetime.strptime(parts[5] + parts[6] + '+00:00',
                                            '%Y%m%d%H%M%S.%f%z'))
        elif len(parts) == 10:
            _id = '_'.join(parts[:6])
            tstamp.append(datetime.strptime(parts[7] + parts[8] + '+00:00',
                                            '%Y%m%d%H%M%S.%f%z'))
        elif len(parts) == 7:
            _id = '_'.join(parts[:3])
            tstamp.append(datetime.strptime(parts[4] + parts[5] + '+00:00',
                                            '%Y%m%d%H%M%S.%f%z'))
        else:
            raise ValueError("Invalid format of data-product name")

        if msm_id is None:
            msm_id = _id
        elif msm_id != _id:
            raise ValueError("Do not combine different kind of measurements")

        dem = DEMio(flname)
        # obtain Science_HK information from header file (ASCII)
        img_hk[ii] = dem.get_sci_hk()
        # get nr_coaddings from file name
        coad_str = [x for x in parts if x.startswith('coad')][0]
        img_hk[ii]['REG_NCOADDFRAMES'] = int(coad_str[-2:])
        # determine exposure time
        t_exp[ii] = dem.exp_time()
        t_frm[ii] = dem.frame_period(int(coad_str[-2:]))
        offset[ii] = dem.offset
        # read image data
        image_list.append(dem.get_data())

    images = np.array(image_list)
    del image_list

    # sort data according to timestamps
    if tstamp != sorted(tstamp):
        print('[WARNING]: do a sort of all the data...')
        indx = sorted(range(n_images), key=tstamp.__getitem__)
        img_hk = img_hk[indx]
        tstamp = [tstamp[k] for k in indx]
        t_exp = t_exp[indx]
        t_frm = t_frm[indx]
        offset = offset[indx]
        temp = temp[indx]
        images = images[indx]

    # convert timestamps to seconds per day
    img_sec = np.empty(n_images, dtype='u4')
    img_subsec = np.empty(n_images, dtype='u2')
    for ii, tval in enumerate(tstamp):
        img_sec[ii] = (tval - EPOCH).total_seconds()
        img_subsec[ii] = tval.microsecond * 2**16 // 1000000

    # generate name of L1A product
    prod_name = spx_product.prod_name(tstamp[0], msm_id=msm_id.strip(' '))
    if args.output is not None:
        dest_dir = Path(args.output)
        if not dest_dir.is_dir():
            dest_dir.mkdir()
        prod_name = str(dest_dir / prod_name)

    # set dimensions of L1A datasets
    if images.ndim == 2:
        n_samples = images.size
    else:
        n_samples = images.shape[1] * images.shape[2]

    dims = {'number_of_images': n_images,
            'samples_per_image': n_samples,
            'SC_records': None,
            'tlm_packets': None,
            'nv': 1}

    # generate L1A product
    with L1Aio(prod_name, dims=dims) as l1a:
        # write image data, detector telemetry and image attributes
        l1a.fill_science(images.reshape(n_images, n_samples), img_hk,
                         np.arange(n_images))
        l1a.fill_time(img_sec, img_subsec, group='image_attributes')

        # Engineering data
        l1a.fill_nomhk(hk_data)
        l1a.fill_time(img_sec, img_subsec, group='engineering_data')

        # Global attributes
        l1a.fill_global_attrs(inflight=False)
        l1a.set_attr('input_files', [Path(x).name for x in args.file_list])

    # Add OGSE and EGSE parameters
    # - Light source
    # - DoLP
    # - AoLP
    # - FOV_begin
    # - FOV_end
    # - Detector illumination
    # - Illumination_level
    # - viewport(s)
    # - Wavelength and signal of data stimulus
    with LV1gse(prod_name) as gse:
        gse.set_attr('origin', 'SPEXone Detector Characterization (SRON)')
        gse.set_attr('measurement', msm_id)
        gse.set_attr('dem_id', args.dem_id)
        if args.reference is not None:
            sec_bgn = int((tstamp[0] - EPOCH).total_seconds())
            sec_end = round((tstamp[-1] - EPOCH).total_seconds())
            with h5py.File(args.reference, 'r') as fid:
                secnd = fid['sec'][:]
                data = fid['amps_nlin'][((secnd >= sec_bgn)
                                         & (secnd <= sec_end))]
            gse.write_reference_signal(data.mean(), data.std())


# --------------------------------------------------
if __name__ == '__main__':
    main()



"""
This file is part of pyspex

https://github.com/rmvanhees/pyspex.git

Python class to access SPEXone Science telemetry data (ApID 0x350)

Copyright (c) 2020-2021 SRON - Netherlands Institute for Space Research
   All Rights Reserved

License:  BSD-3-Clause
"""
import numpy as np


# - global parameters -------------------


# - class TMscience -------------------------
class TMscience:
    """
    Access/convert parameters of a SPEXone telemetry Science packet

    Methods
    -------
    get(key)
       Return (raw) SPEXone telemetry Science parameter.
    number_channels
       Return number of LVDS channels used.
    binning_table_id
       Return the binning table identifier (zero for full-frame images).
    lvds_clock
       Returns flag for LVDS clock: False: disabled & True: enabled.
    offset
       Returns digital offset including ADC offset [counts].
    pga_gain
       Returns PGA gain [Volt].
    exp_time
       Returns pixel exposure time [master clock periods].
    fot_time
       Returns frame overhead time [master clock periods].
    rot_time
       Returns image read-out time [master clock periods].
    frame_period
       Returns frame period [master clock periods].
    pll_control
       Returns raw PLL control parameters: pll_range, pll_out_fre, pll_div.
    exp_control
       Returns raw exposure time parameters: inte_sync, exp_dual, exp_ext.
    """
    def __init__(self, tm_science):
        """
        Initialize class TMscience

        Parameters
        tm_science :  ndarray
           SPEXone telemetry Science data
        """
        self.__tm = tm_science

    def get(self, key: str):
        """
        Return (raw) Science telemetry parameter
        """
        return self.__tm[key] if key in self.__tm.dtype.names else None

    @property
    def binning_table_id(self) -> int:
        """
        Return the binning table identifier (zero for full-frame images)

        Notes
        -----
        The CCSDS data may hold data collected with different MPS, but
        only when the size of the detector data does not change. Therefore,
        a mix of Science mode and Full-frame data should not occur.

        v126: Sometimes the MPS information is not updated for the first
              images. We try to fix this and warn the user.
        v129: REG_BINNING_TABLE_START is stored in BE instead of LE
        """
        full_frame = np.unique(self.__tm['REG_FULL_FRAME'])
        if len(full_frame) > 1:
            print('[WARNING]: value of REG_FULL_FRAME not unique')
            print(self.__tm['REG_FULL_FRAME'])
        full_frame = self.__tm['REG_FULL_FRAME'][-1]

        cmv_outputmode = np.unique(self.__tm['REG_CMV_OUTPUTMODE'])
        if len(cmv_outputmode) > 1:
            print('[WARNING]: value of REG_CMV_OUTPUTMODE not unique')
            print(self.__tm['REG_CMV_OUTPUTMODE'])
        cmv_outputmode = self.__tm['REG_CMV_OUTPUTMODE'][-1]

        if full_frame == 1:
            if cmv_outputmode != 3:
                raise KeyError('Diagnostic mode with REG_CMV_OUTPMODE != 3')
            return np.zeros(len(self.__tm), dtype='i1')

        if full_frame == 2:
            if cmv_outputmode != 1:
                raise KeyError('Science mode with REG_CMV_OUTPUTMODE != 1')
            bin_tbl_start = self.__tm['REG_BINNING_TABLE_START']
            indx0 = (self.__tm['REG_FULL_FRAME'] != 2).nonzero()[0]
            if indx0.size > 0:
                indx2 = (self.__tm['REG_FULL_FRAME'] == 2).nonzero()[0]
                bin_tbl_start[indx0] = bin_tbl_start[indx2[0]]
            res = 1 + (bin_tbl_start - 0x80000000) // 0x400000
            return res & 0xFF

        raise KeyError('REG_FULL_FRAME not equal to 1 or 2')

    @property
    def number_channels(self) -> int:
        """
        Return number of LVDS channels used
        """
        return 2 ** (4 - (self.__tm['DET_OUTMODE'] & 0x3))

    @property
    def lvds_clock(self) -> bool:
        """
        Returns flag for LVDS clock: False: disabled & True: enabled
        """
        return ((self.__tm['DET_PLLENA'] & 0x3) == 0
                and (self.__tm['DET_PLLBYP'] & 0x3) != 0
                and (self.__tm['DET_CHENA'] & 0x40000) != 0)

    @property
    def offset(self) -> int:
        """
        Returns digital offset including ADC offset
        """
        buff = self.__tm['DET_OFFSET'].astype('i4')
        if np.isscalar(buff):
            if buff >= 8192:
                buff -= 16384
        else:
            buff[buff >= 8192] -= 16384

        return buff + 70

    @property
    def pga_gain(self) -> float:
        """
        Returns PGA gain [Volt]
        """
        # need first bit of address 121
        reg_pgagainfactor = self.__tm['DET_BLACKCOL'] & 0x1

        reg_pgagain = self.__tm['DET_PGAGAIN']

        return (1 + 0.2 * reg_pgagain) * 2 ** reg_pgagainfactor

    @property
    def exp_time(self) -> float:
        """
        Returns pixel exposure time [master clock periods]
        """
        return 129 * (0.43 * self.__tm['DET_FOTLEN']
                      + self.__tm['DET_EXPTIME'])

    @property
    def fot_time(self) -> int:
        """
        Returns frame overhead time [master clock periods]
        """
        return 129 * (self.__tm['DET_FOTLEN']
                      + 2 * (16 // self.number_channels))

    @property
    def rot_time(self) -> int:
        """
        Returns image read-out time [master clock periods]
        """
        return 129 * (16 // self.number_channels) * self.__tm['DET_NUMLINES']

    @property
    def frame_period(self) -> float:
        """
        Returns frame period [master clock periods]
        """
        return 2.38e-7 + (self.__tm['REG_NCOADDFRAMES']
                          * (self.exp_time + self.fot_time + self.rot_time))

    @property
    def pll_control(self) -> tuple:
        """
        Returns raw PLL control parameters: pll_range, pll_out_fre, pll_div

        Notes
        -----
        PLL_range:    bits [7], valid values: 0 or 1
        PLL_out_fre:  bits [4:7], valid values:  0, 1, 2 or 5
        PLL_div:      bits [0:3], valid values 9 (10-bit) or 11 (12-bit)

        Other PLL registers are: PLL_enable, PLL_in_fre, PLL_bypass, PLL_load
        """
        pll_div = self.__tm['DET_PLLRATE'] & 0xF             # bit [0:4]
        pll_out_fre = (self.__tm['DET_PLLRATE'] >> 4) & 0x7  # bit [4:7]
        pll_range = (self.__tm['DET_PLLRATE'] >> 7)          # bit [7]

        return (pll_range, pll_out_fre, pll_div)

    @property
    def exp_control(self) -> tuple:
        """
        Returns raw exposure time parameters: inte_sync, exp_dual, exp_ext
        """
        inte_sync = (self.__tm['INTE_SYNC'] >> 2) & 0x1
        exp_dual = (self.__tm['INTE_SYNC'] >> 1) & 0x1
        exp_ext = self.__tm['INTE_SYNC'] & 0x1

        return (inte_sync, exp_dual, exp_ext)



"""
This file is part of pyspex

https://github.com/rmvanhees/pyspex.git

Classes to read SPEXone Level-0 packages are ScienceCCSDS and TmTcCCSDS for
resp. Science data and telemetry data. Both classes can be used to read data
in pure CCSDS format and ST3 format. The latter wil generated after integration
of SPEXone on the PACE platform.

Both classes have a much higher performance than obsolete class CCSDSio.

ToDo
----
- Remove the 16-bit ITOS header from the ST3-format hdr_dtype definition
- Add data sanity checks

Copyright (c) 2022 SRON - Netherlands Institute for Space Research
   All Rights Reserved

License:  BSD-3-Clause
"""
import numpy as np

from pyspex.lib.tmtc_def import tmtc_dtype

# - global parameters ------------------------------
FULLFRAME_BYTES = 2 * 2048 * 2048


# - local functions --------------------------------
def hdr_dtype(st3_format=True) -> np.dtype:
    """
    Returns numpy dtype for packet header of data in CCSDS or ST3 format
    """
    if st3_format:
        return np.dtype([('itos_hdr', '>u2', (8,)),
                         ('spacewire', '>u2'),
                         ('type', '>u2'),
                         ('sequence', '>u2'),
                         ('length', '>u2'),
                         ('tai_sec', '>u4'),
                         ('sub_sec', '>u2')])

    return np.dtype([('type', '>u2'),
                     ('sequence', '>u2'),
                     ('length', '>u2'),
                     ('tai_sec', '>u4'),
                     ('sub_sec', '>u2')])


def ap_id(hdr: np.ndarray) -> int:
    """
    Returns Telemetry APID, the range 0x320 to 0x351 is available to SPEXone

    The following values are recognized:
        0x350 : Science
        0x320 : NomHk
        0x322 : DemHk
        0x331 : TcAccept
        0x332 : TcReject
        0x333 : TcExecute
        0x334 : TcFail
    """
    return hdr['type'] & 0x7FF


def grouping_flag(hdr: np.ndarray) -> int:
    """
    Returns grouping flag

    The 2-byte flag is encoded as follows:
        00 continuation packet-data segment
        01 first packet-data segment
        10 last packet-data segment
        11 packet-data unsegmented
    """
    return (hdr['sequence'] >> 14) & 0x3


def sequence(hdr: np.ndarray) -> int:
    """
    Returns sequence counter, rollover to zero at 0x3FFF
    """
    return hdr['sequence'] & 0x3FFF


def packet_length(hdr: np.ndarray) -> int:
    """
    Returns size of secondary header + user data - 1 in bytes

    Notes
    -----
    We always read the primary header and secondary header at once.

    Value range: 7 - 16375
    """
    return hdr['length']


def split_ccsds(ccsds_data: bytes) -> tuple:
    """
    Split Science and Telemetry data and store each in tuples (ST3 only)
    """
    _hdr_dtype = hdr_dtype(True)
    _hdr_bytes = _hdr_dtype.itemsize - 5

    offs = 0
    ccsds_sci = ()
    ccsds_hk = ()
    while offs < len(ccsds_data):
        hdr = np.frombuffer(ccsds_data, count=1, offset=offs,
                            dtype=_hdr_dtype)[0]
        nbytes = hdr['length'] + _hdr_bytes
        if ap_id(hdr) == 0x350:
            ccsds_sci += (ccsds_data[offs:offs+nbytes],)
        else:
            ccsds_hk += (ccsds_data[offs:offs+nbytes],)
        offs += nbytes

    return (ccsds_sci, ccsds_hk)


def select_science(ccsds_data: bytes) -> tuple:
    """
    Select Science CCSDS packages
    """
    _hdr_dtype = hdr_dtype(False)
    _hdr_bytes = _hdr_dtype.itemsize - 5

    offs = 0
    ccsds_sci = ()
    while offs < len(ccsds_data):
        hdr = np.frombuffer(ccsds_data, count=1, offset=offs,
                            dtype=_hdr_dtype)[0]
        nbytes = hdr['length'] + _hdr_bytes
        if ap_id(hdr) == 0x350:
            ccsds_sci += (ccsds_data[offs:offs+nbytes],)
        offs += nbytes

    return ccsds_sci


def select_hk(ccsds_data: bytes) -> tuple:
    """
    Select telemetry CCSDS packages
    """
    _hdr_dtype = hdr_dtype(False)
    _hdr_bytes = _hdr_dtype.itemsize - 5

    offs = 0
    ccsds_hk = ()
    while offs < len(ccsds_data):
        hdr = np.frombuffer(ccsds_data, count=1, offset=offs,
                            dtype=_hdr_dtype)[0]
        nbytes = hdr['length'] + _hdr_bytes
        if ap_id(hdr) != 0x350:
            ccsds_hk += (ccsds_data[offs:offs+nbytes],)
        offs += nbytes

    return ccsds_hk


# - class definitions ------------------------------
class ScienceCCSDS():
    """
    Read SPEXone Science data
    """
    APID = 0x350

    def __init__(self, ccsds_data: tuple, verbose=False):
        """
        Parameters
        ----------
        ccsds_data :  tuple
           SPEXone science data packages in CCSDS or ST3 format
        verbose :  bool, optional
           Be verbose
        """
        self.verbose = verbose
        self.ccsds_data = ccsds_data
        self._offs = 0

        # check if data-format is CCSDS or ST3
        if self.is_ccsds:
            st3_format = False
            self.hdr_dtype = hdr_dtype(st3_format)
        elif self.is_st3:
            st3_format = True
            self.hdr_dtype = hdr_dtype(st3_format)
        else:
            raise ValueError('format error')
        self.icutm_dtype = np.dtype([('tai_sec', '>u4'),
                                     ('sub_sec', '>u2')])
        if verbose:
            print('[INFO]: Science data is in'
                  f' {"ST3" if st3_format else "CCSDS"} format')

        # start processing at the first package with grouping flag in (1, 3)
        indx = self.find_first_segment()
        if indx > 0:
            self.ccsds_data = self.ccsds_data[indx:]

    @property
    def scihk_dtype(self) -> np.dtype:
        """
        Returns numpy dtype for Science housekeeping data
        """
        return tmtc_dtype(self.APID)

    @property
    def is_ccsds(self) -> bool:
        """
        Returns True if data is in CCSDS format
        """
        hdr = np.frombuffer(self.ccsds_data[0], count=1, offset=0,
                            dtype=hdr_dtype(st3_format=False))
        if (hdr['type'] >> 11) == 1 and ap_id(hdr) == self.APID:
            return True

        return False

    @property
    def is_st3(self) -> bool:
        """
        Returns True if data is in ST3 format
        """
        hdr = np.frombuffer(self.ccsds_data[0], count=1, offset=0,
                            dtype=hdr_dtype(st3_format=True))
        if (hdr['type'] >> 11) == 1 and ap_id(hdr) == self.APID:
            return True

        return False

    def find_first_segment(self) -> int:
        """
        Find byte offset to first segment with grouping flag equal 1 or 3
        """
        indx = 0
        for segment in self.ccsds_data:
            hdr = np.frombuffer(segment, count=1, offset=0,
                                dtype=self.hdr_dtype)[0]
            if ap_id(hdr) == self.APID and grouping_flag(hdr) in (1, 3):
                break
            indx += 1

        if self.verbose:
            print(f'[INFO]: found first valid segment at {indx}')
        return indx

    # def check(self) -> None:
    #    """
    #    Perform sanity checks, reading packet-headers only
    #    """
    #    return

    @staticmethod
    def _fix_hk24_(sci_hk):
        """
        Correct 32-bit integers in the Science HK which originate from
        24-bit integers in the detector register values

        In addition:
         - copy the first 4 bytes of DET_CHENA to DET_ILVDS
         - parameter 'REG_BINNING_TABLE_START' was writen in little-endian
        """
        res = sci_hk.copy()
        if sci_hk['ICUSWVER'] < 0x129:
            res['REG_BINNING_TABLE_START'] = \
                sci_hk['REG_BINNING_TABLE_START'].byteswap()

        res['DET_ILVDS'] = sci_hk['DET_CHENA'] & 0xf
        for key in ['TS1_DEM_N_T', 'TS2_HOUSING_N_T', 'TS3_RADIATOR_N_T',
                    'TS4_DEM_R_T', 'TS5_HOUSING_R_T', 'TS6_RADIATOR_R_T',
                    'LED1_ANODE_V', 'LED1_CATH_V', 'LED1_I',
                    'LED2_ANODE_V', 'LED2_CATH_V', 'LED2_I',
                    'ADC1_VCC', 'ADC1_REF', 'ADC1_T',
                    'ADC2_VCC', 'ADC2_REF', 'ADC2_T',
                    'DET_EXPTIME', 'DET_EXPSTEP', 'DET_KP1',
                    'DET_KP2', 'DET_EXPTIME2', 'DET_EXPSTEP2',
                    'DET_CHENA']:
            res[key] = sci_hk[key] >> 8
        return res

    def dump(self, flname: str) -> None:
        """
        Dump header information of all data segments

        Parameters
        ----------
        flname :  str
          Name of the ASCII file to be overwritten with a data dump
        """
        # pylint: disable=consider-using-with
        fp = open(flname, 'w', encoding='ascii')
        fp.write('APID Grouping Counter Length'
                 ' ICUSWVER MPS_ID  IMRLEN     ICU_SEC ICU_SUBSEC\n')

        # process all data
        for segment in self.ccsds_data:
            try:
                hdr = np.frombuffer(segment, count=1, offset=0,
                                    dtype=self.hdr_dtype)[0]
                offs = self.hdr_dtype.itemsize
                if grouping_flag(hdr) in (1, 3):
                    sci_hk = np.frombuffer(segment, count=1, offset=offs,
                                           dtype=self.scihk_dtype)[0]
                    offs += self.scihk_dtype.itemsize
                    icu_tm = np.frombuffer(segment, count=1, offset=offs,
                                           dtype=self.icutm_dtype)[0]
                    offs += self.icutm_dtype.itemsize
                    fp.write(f"{ap_id(hdr):4x} {grouping_flag(hdr):8d}"
                             f" {sequence(hdr):7d} {packet_length(hdr):6d}"
                             f" {sci_hk['ICUSWVER']:8x} {sci_hk['MPS_ID']:6d}"
                             f" {sci_hk['IMRLEN']:7d} {icu_tm['tai_sec']:11d}"
                             f" {icu_tm['sub_sec']:10d}\n")
                else:
                    fp.write(f'{ap_id(hdr):4x} {grouping_flag(hdr):8d}'
                             f' {sequence(hdr):7d} {packet_length(hdr):6d}\n')
            except EOFError:
                pass
            except Exception as exc:
                raise RuntimeError('an exception occured during read') from exc

        fp.close()

    def read(self, select='all') -> tuple:
        """
        Read SPEXone Science data into a tuple of numpy arrays
        """
        diff_bytes = self.icutm_dtype.itemsize - 1
        diff_bytes1 = \
            2 * self.icutm_dtype.itemsize + self.scihk_dtype.itemsize - 1

        buff = np.empty(1, dtype=np.dtype([('hdr', self.hdr_dtype),
                                           ('hk', self.scihk_dtype),
                                           ('icu_tm', self.icutm_dtype),
                                           ('frame', 'O')]))
        # process all data
        frame = ()
        packets = ()
        for segment in self.ccsds_data:
            try:
                hdr = np.frombuffer(segment, count=1, offset=0,
                                    dtype=self.hdr_dtype)[0]
                offs = self.hdr_dtype.itemsize
                if grouping_flag(hdr) == 1:
                    buff['hdr'] = hdr
                    buff['hk'] = self._fix_hk24_(
                        np.frombuffer(segment, count=1, offset=offs,
                                      dtype=self.scihk_dtype)[0])
                    offs += self.scihk_dtype.itemsize
                    buff['icu_tm'] = np.frombuffer(segment, count=1,
                                                   offset=offs,
                                                   dtype=self.icutm_dtype)[0]
                    offs += self.icutm_dtype.itemsize
                    frame = ()
                    nbytes = hdr['length'] - diff_bytes1
                else:
                    nbytes = hdr['length'] - diff_bytes
                frame += (np.frombuffer(segment, dtype='>u2',
                                        count=nbytes // 2, offset=offs),)
                if grouping_flag(hdr) == 2:
                    buff['frame'][0] = np.concatenate(frame)
                    if select == 'all':
                        packets += (buff.copy(),)
                    elif (select == 'binned'
                          and buff['hk']['IMRLEN'][0] < FULLFRAME_BYTES):
                        packets += (buff.copy(),)
                    elif (select == 'fullFrame'
                          and buff['hk']['IMRLEN'][0] == FULLFRAME_BYTES):
                        packets += (buff.copy(),)
            except EOFError:
                pass
            except Exception as exc:
                raise RuntimeError('an exception occured') from exc

        return packets


#--------------------------------------------------
class TmTcCCSDS():
    """
    Read SPEXone telemetry data
    """
    VALID_APID = [0x320, 0x322, 0x331, 0x332, 0x333, 0x334]

    def __init__(self, ccsds_data: tuple, verbose=False):
        """
        Parameters
        ----------
        ccsds_data :  tuple
           SPEXone telemetry data packages in CCSDS or ST3 format
        verbose :  bool, optional
           Be verbose
        """
        self.verbose = verbose
        self.ccsds_data = ccsds_data
        self._offs = 0

        # check if data-format is CCSDS or ST3
        if self.is_ccsds:
            st3_format = False
            self.hdr_dtype = hdr_dtype(st3_format)
        elif self.is_st3:
            st3_format = True
            self.hdr_dtype = hdr_dtype(st3_format)
        else:
            raise ValueError('format error')
        if verbose:
            print('[INFO]: TmTc data is in'
                  f' {"ST3" if st3_format else "CCSDS"} format')

    @property
    def is_ccsds(self) -> bool:
        """
        Returns True if data is in CCSDS format
        """
        hdr = np.frombuffer(self.ccsds_data[0], count=1, offset=0,
                            dtype=hdr_dtype(st3_format=False))
        if (hdr['type'] >> 11) == 1 and ap_id(hdr) in self.VALID_APID:
            return True

        return False

    @property
    def is_st3(self) -> bool:
        """
        Returns True if data is in ST3 format
        """
        hdr = np.frombuffer(self.ccsds_data[0], count=1, offset=0,
                            dtype=hdr_dtype(st3_format=True))
        if (hdr['type'] >> 11) == 1 and ap_id(hdr) in self.VALID_APID:
            return True

        return False

    @property
    def dtype_dict(self) -> dict:
        """
        Return dictionary with numpy dtype definitions for telemetry packages
        """
        return {0x320: np.dtype([('hdr', self.hdr_dtype),
                                 ('hk', tmtc_dtype(0x320))]),
                0x322: np.dtype([('hdr', self.hdr_dtype),
                                 ('hk', tmtc_dtype(0x322))]),
                0x331: np.dtype([('hdr', self.hdr_dtype),
                                 ('TcPacketId', '>u2'),
                                 ('TcSeqControl', '>u2')]),
                0x332: np.dtype([('hdr', self.hdr_dtype),
                                 ('TcPacketId', '>u2'),
                                 ('TcSeqControl', '>u2'),
                                 ('TcErrorCode', '>u2'),
                                 ('RejectParameter1', '>u2'),
                                 ('RejectParameter2', '>u2')]),
                0x333: np.dtype([('hdr', self.hdr_dtype),
                                 ('TcPacketId', '>u2'),
                                 ('TcSeqControl', '>u2')]),
                0x334: np.dtype([('hdr', self.hdr_dtype),
                                 ('TcPacketId', '>u2'),
                                 ('TcSeqControl', '>u2'),
                                 ('TcErrorCode', '>u2'),
                                 ('FailParameter1', '>u2'),
                                 ('FailParameter2', '>u2')])}

    def dump(self, flname: str) -> None:
        """
        Dump header information of all data segments

        Parameters
        ----------
        flname :  str
          Name of the ASCII file to be overwritten with a data dump
        """
        # pylint: disable=consider-using-with
        fp = open(flname, 'w', encoding='ascii')
        fp.write('APID Grouping Counter Length     TAI_SEC    SUB_SEC'
                 ' ICUSWVER MPS_ID TcSeqControl TcErrorCode\n')

        for segment in self.ccsds_data:
            try:
                hdr = np.frombuffer(segment, count=1, offset=0,
                                    dtype=self.hdr_dtype)[0]
                offs = self.hdr_dtype.itemsize
                if ap_id(hdr) == 0x320:
                    mon_hk = np.frombuffer(segment, count=1, offset=offs,
                                           dtype=tmtc_dtype(0x320))[0]
                    fp.write(f"{ap_id(hdr):4x} {grouping_flag(hdr):8d}"
                             f" {sequence(hdr):7d} {packet_length(hdr):6d}"
                             f" {hdr['tai_sec']:11d} {hdr['sub_sec']:10d}"
                             f" {mon_hk['ICUSWVER']:8x}"
                             f" {mon_hk['MPS_ID']:6d}\n")
                elif ap_id(hdr) in (0x332, 0x334):
                    tc_dtype = self.dtype_dict[ap_id(hdr)]
                    tc_val = np.frombuffer(segment, count=1, offset=0,
                                           dtype=tc_dtype)[0]
                    fp.write(f"{ap_id(hdr):4x} {grouping_flag(hdr):8d}"
                             f" {sequence(hdr):7d} {packet_length(hdr):6d}"
                             f" {hdr['tai_sec']:11d} {hdr['sub_sec']:10d}"
                             f" {-1:8x} {-1:6d} {tc_val['TcSeqControl']:12d}"
                             f" {bin(tc_val['TcErrorCode'])}\n")
                else:
                    fp.write(f"{ap_id(hdr):4x} {grouping_flag(hdr):8d}"
                             f" {sequence(hdr):7d} {packet_length(hdr):6d}"
                             f" {hdr['tai_sec']:11d} {hdr['sub_sec']:10d}\n")
            except Exception as exc:
                raise RuntimeError('an exception occured during read') from exc

        fp.close()

    def read(self):
        """
        Read SPEXone telemetry data into a tuple of numpy arrays
        """
        # reset private class attribute
        packets = ()
        for segment in self.ccsds_data:
            try:
                hdr = np.frombuffer(segment, count=1, offset=0,
                                    dtype=self.hdr_dtype)[0]
                tm_dtype = self.dtype_dict.get(ap_id(hdr), None)
                if tm_dtype is not None:
                    packets += (np.frombuffer(segment, count=1, offset=0,
                                              dtype=tm_dtype),)
            except Exception as exc:
                raise RuntimeError('an exception occured during read') from exc

        return packets



"""
This file is part of pyspex

https://github.com/rmvanhees/pyspex.git

Class to read SPEXone ICU packages (version 2)

Copyright (c) 2019-2022 SRON - Netherlands Institute for Space Research
   All Rights Reserved

License:  BSD-3-Clause
"""
from pathlib import Path

import numpy as np

from pyspex.lib.tmtc_def import tmtc_dtype

# - global parameters ------------------------------
# Define parameters of Primary header
#  - Packet type     (3 bits): Version No.
#                              Indicates this is a CCSDS version 1 packet
#                     (1 bit): Type indicator
#                              Indicates this is a telemetery packet
#                     (1 bit): Secondary flag
#                              Indicate presence of Secondary header
#                   (11 bits): ApID
#                              SPEXone ApID [0x320 - 0x351] or 2047
#
#  - Packet Sequence (2 bits): Grouping flag
#                              00 continuation packet-data segment
#                              01 first packet-data segment
#                              10 last packet-data segment
#                              11 packet-data unsegmented
#                   (14 bits): Counter per ApID, rollover to 0 at 0x3FFF
#  - Packet length  (16 bits): size of packet data in bytes (always odd)
#                              (secondary header + User data) - 1
#  - Packet timestamp:         Secondary header
#                   (32 bits): seconds (1-1-1970 UTC)
#                   (16 bits): sub-seconds (1/2 ** 16)

HDR_DTYPE = np.dtype([
    ('type', '>u2'),
    ('sequence', '>u2'),
    ('length', '>u2'),
    ('tai_sec', '>u4'),
    ('sub_sec', '>u2')
])

TIME_DTYPE = np.dtype([
    ('tai_sec', '>u4'),
    ('sub_sec', '>u2')
])

SCIHK_DTYPE = tmtc_dtype(0x350)

# - Error messages ------------------------
MSG_SKIP_FRAME = "[WARNING]: rejected a frame because it's incomplete"
MSG_INVALID_APID = \
    '[WARNING]: found one or more telemetry packages with an invalid APID'
MSG_CORRUPT_APID = 'corrupted segements - detected APID 1 after <> 2'
MSG_CORRUPT_FRAME = 'corrupted segements - previous frame not closed'


# - class CCSDSio -------------------------
class CCSDSio:
    """
    Read SPEXone telemetry packets.

    Attributes
    ----------
    found_invalid_apid :  bool
    file_list :  iter
    fp :  file pointer

    Methods
    -------
    version_no
       Returns CCSDS version number.
    type_indicator
       Returns type of telemetry packet.
    secnd_hdr_flag
       Returns flag indicating presence of a secondary header.
    ap_id
       Returns SPEXone ApID.
    grouping_flag
       Returns grouping flag.
    sequence_count
       Returns sequence counter, rollover to zero at 0x3FFF.
    packet_length
       Returns size of packet data in bytes.

    close()
       Close resources.
    open_next_file()
       Open next file from file_list.
    fix_dem_hk24(sci_hk)
       Correct 32-bit integers in the DemHK which originate from
       24-bit integers in the detector register values.
    fix_sci_hk24(sci_hk)
       Correct 32-bit integers in the Science_HK which originate from
       24-bit integers in the detector register values.
    read_packet()
       Read next telemetry packet.
    select_tm(packets, ap_id)
       Select NomHK telemetry packages
    science_tm(packets_in)
       Combine segmented Science telemetry packages.

    Notes
    -----
    The formats of the PACE telemetry packets are following the standards:
    CCSDS-131.0-B-3, CCSDS-132.0-B-2 and CCSDS-133.0-B-1.

    This module is currenty restriced to telementry packets with APID:
    0x350 (Science), 0x320 (NomHK) and 0x322 (DemHK).

    A telemtry packet consist of a PRIMARY HEADER, SECONDARY HEADER
    (consist of a timestamp) and USER DATA with the actual telemetry
    packet data.

    Doc: TMTC handbook (SPX1-TN-005), issue 12, 15-May-2020

    Examples
    --------
    >>> packets = ()
    >>> with CCSDSio(file_list) as ccsds:
    >>>     while True:
    >>>         # read one telemetry packet at a time
    >>>         packet = ccsds.read_packet()
    >>>         if packet is None:
    >>>             # now we have read all files
    >>>             break
    >>>
    >>>         packets += (packet[0],)
    >>>
    >>>     # combine segmented Science packages
    >>>     science_tm = ccsds.group_tm(packets)
    >>>     # now you may want to collect the engineering packages
    """
    def __init__(self, file_list: str) -> None:
        """
        Initialize access to a SPEXone Level-0 product (CCSDS format)

        Parameters
        ----------
        file_list: list of strings
           list of file-names, where each file contains parts of a measurement

        Notes
        -----
        The files with science and telemetry data needs to be in chronological
        order. However, you may mix science and housekeeping data as long as
        science data are chronological and housekeeping data are chronological.
        """
        # initialize class attributes
        self.__hdr = None
        self.found_invalid_apid = False
        self.file_list = iter(file_list)
        self.fp = None

        self.open_next_file()

    def __repr__(self) -> str:
        return (f'{self.version_no:03d} 0x{self.ap_id:x}'
                f' {self.grouping_flag} {self.sequence_count:5d}'
                f' {self.packet_length:5d} {self.fp.tell():9d}')

    def __iter__(self):
        for attr in sorted(self.__dict__):
            if not attr.startswith("__"):
                yield attr

    def __enter__(self):
        """
        method called to initiate the context manager
        """
        return self

    def __exit__(self, exc_type, exc_value, traceback) -> bool:
        """
        method called when exiting the context manager
        """
        self.close()
        return False  # any exception is raised by the with statement.

    def close(self) -> None:
        """
        Close resources
        """
        if self.fp is not None:
            if self.found_invalid_apid:
                print(MSG_INVALID_APID)
            self.found_invalid_apid = False
            self.fp.close()

    # ---------- define some class properties ----------
    @property
    def version_no(self) -> int:
        """
        Returns CCSDS version number
        """
        if self.__hdr is None:
            return None

        return (self.__hdr['type'] >> 13) & 0x7

    @property
    def type_indicator(self) -> int:
        """
        Returns type of telemetry packet
        """
        if self.__hdr is None:
            return None

        return (self.__hdr['type'] >> 12) & 0x1

    @property
    def secnd_hdr_flag(self) -> bool:
        """
        Returns flag indicating presence of a secondary header
        """
        if self.__hdr is None:
            return None

        return (self.__hdr['type'] >> 11) & 0x1

    @property
    def ap_id(self) -> int:
        """
        Returns SPEXone ApID
        """
        if self.__hdr is None:
            return None

        return self.__hdr['type'] & 0x7FF

    @property
    def grouping_flag(self) -> int:
        """
        Returns grouping flag

        Possible values:
          00 continuation packet-data segment
          01 first packet-data segment
          10 last packet-data segment
          11 packet-data unsegmented
        """
        if self.__hdr is None:
            return None

        return (self.__hdr['sequence'] >> 14) & 0x3

    @property
    def sequence_count(self) -> int:
        """
        Returns sequence counter, rollover to zero at 0x3FFF
        """
        if self.__hdr is None:
            return None

        return self.__hdr['sequence'] & 0x3FFF

    @property
    def packet_length(self) -> int:
        """
        Returns size of packet data in bytes

        Notes
        -----
        Value equals secondary header + user data (always odd)
        """
        if self.__hdr is None:
            return None

        return self.__hdr['length']

    # ---------- define empty telemetry packet ----------
    def open_next_file(self) -> None:
        """
        Open next file from file_list
        """
        flname = next(self.file_list)
        if not Path(flname).is_file():
            raise FileNotFoundError(f'{flname} does not exist')

        self.close()
        # pylint: disable=consider-using-with
        self.fp = open(flname, 'rb')

    @staticmethod
    def fix_dem_hk24(dem_hk):
        """
        Correct 32-bit integers in the DemHK which originate from
        24-bit integers in the detector register values
        """
        for key in ['DET_EXPTIME', 'DET_EXPSTEP',
                    'DET_KP1', 'DET_KP2', 'DET_EXPTIME2', 'DET_EXPSTEP2']:
            dem_hk[key] = dem_hk[key] >> 8

        return dem_hk

    @staticmethod
    def fix_sci_hk24(sci_hk):
        """
        Correct 32-bit integers in the Science HK which originate from
        24-bit integers in the detector register values

        In addition:
         - copy the first 4 bytes of DET_CHENA to DET_ILVDS
         - parameter 'REG_BINNING_TABLE_START' was writen in little-endian
        """
        if np.all(sci_hk['ICUSWVER'] < 0x129):
            key = 'REG_BINNING_TABLE_START'
            sci_hk[key] = np.ndarray(shape=sci_hk.shape,
                                     dtype='<u4',
                                     buffer=sci_hk[key])

        sci_hk['DET_ILVDS'] = sci_hk['DET_CHENA'] & 0xf

        for key in ['TS1_DEM_N_T', 'TS2_HOUSING_N_T', 'TS3_RADIATOR_N_T',
                    'TS4_DEM_R_T', 'TS5_HOUSING_R_T', 'TS6_RADIATOR_R_T',
                    'LED1_ANODE_V', 'LED1_CATH_V', 'LED1_I',
                    'LED2_ANODE_V', 'LED2_CATH_V', 'LED2_I',
                    'ADC1_VCC', 'ADC1_REF', 'ADC1_T',
                    'ADC2_VCC', 'ADC2_REF', 'ADC2_T',
                    'DET_EXPTIME', 'DET_EXPSTEP', 'DET_KP1',
                    'DET_KP2', 'DET_EXPTIME2', 'DET_EXPSTEP2',
                    'DET_CHENA']:
            sci_hk[key] = sci_hk[key] >> 8

        return sci_hk

    def __rd_science(self, hdr) -> np.ndarray:
        """
        Read Science telemetry packet
        """
        num_bytes = self.packet_length - TIME_DTYPE.itemsize + 1
        packet = np.empty(1, dtype=np.dtype([
            ('packet_header', HDR_DTYPE),
            ('science_hk', SCIHK_DTYPE),
            ('icu_time', TIME_DTYPE),
            ('image_data', 'O')]))
        packet['packet_header'] = hdr

        # first segement or unsegmented data packet provides Science_HK
        if self.grouping_flag in (1, 3):
            packet['science_hk'] = self.fix_sci_hk24(
                np.fromfile(self.fp, count=1, dtype=SCIHK_DTYPE))
            num_bytes -= SCIHK_DTYPE.itemsize
            packet['icu_time'] = np.fromfile(self.fp, count=1,
                                             dtype=TIME_DTYPE)
            num_bytes -= TIME_DTYPE.itemsize

        # read detector image data
        packet['image_data'][0] = np.fromfile(self.fp, dtype='>u2',
                                              count=num_bytes // 2)
        return packet

    def __rd_nomhk(self, hdr) -> np.ndarray:
        """
        Read NomHK telemetry packet
        """
        packet = np.empty(1, dtype=np.dtype([
            ('packet_header', HDR_DTYPE),
            ('nominal_hk', tmtc_dtype(0x320))]))
        packet['packet_header'] = hdr
        packet['nominal_hk'] = np.fromfile(self.fp, count=1,
                                           dtype=tmtc_dtype(0x320))
        return packet

    def __rd_demhk(self, hdr) -> np.ndarray:
        """
        Read DemHK telemetry packet
        """
        packet = np.empty(1, dtype=np.dtype([
            ('packet_header', HDR_DTYPE),
            ('detector_hk', tmtc_dtype(0x322))]))
        packet['packet_header'] = hdr
        packet['detector_hk'] = self.fix_dem_hk24(
            np.fromfile(self.fp, count=1, dtype=tmtc_dtype(0x322)))
        return packet

    def __rd_tc_accept(self, _) -> np.ndarray:
        """
        Read/dump TcAccept packet
        """
        self.fp.seek(-1 * HDR_DTYPE.itemsize, 1)
        packet = np.fromfile(self.fp, count=1, dtype=np.dtype([
            ('packet_header', HDR_DTYPE),
            ('TcPacketId', '>u2'),
            ('TcSeqControl', '>u2')]))
        print('[TcAccept]: ', self, packet['TcPacketId'][0],
              packet['TcSeqControl'][0])
        return packet

    def __rd_tc_execute(self, _) -> np.ndarray:
        """
        Read/dump TcExecute packet
        """
        self.fp.seek(-1 * HDR_DTYPE.itemsize, 1)
        packet = np.fromfile(self.fp, count=1, dtype=np.dtype([
            ('packet_header', HDR_DTYPE),
            ('TcPacketId', '>u2'),
            ('TcSeqControl', '>u2')]))
        print('[TcExecute]:', self, packet['TcPacketId'][0],
              packet['TcSeqControl'][0])
        return packet

    def __rd_tc_fail(self, _) -> np.ndarray:
        """
        Read/dump TcFail packet
        """
        self.fp.seek(-1 * HDR_DTYPE.itemsize, 1)
        packet = np.fromfile(self.fp, count=1, dtype=np.dtype([
            ('packet_header', HDR_DTYPE),
            ('TcPacketId', '>u2'),
            ('TcSeqControl', '>u2'),
            ('TcFailCode', '>u2'),
            ('FailParameter1', '>u2'),
            ('FailParameter2', '>u2')]))
        print('[TcFail]:   ', self, packet['TcPacketId'][0],
              packet['TcSeqControl'][0], bin(packet['TcFailCode'][0]),
              packet['FailParameter1'][0], packet['FailParameter2'][0]
              )
        return packet

    def __rd_tc_reject(self, _) -> np.ndarray:
        """
        Read/dump TcReject packet
        """
        self.fp.seek(-1 * HDR_DTYPE.itemsize, 1)
        packet = np.fromfile(self.fp, count=1, dtype=np.dtype([
            ('packet_header', HDR_DTYPE),
            ('TcPacketId', '>u2'),
            ('TcSeqControl', '>u2'),
            ('TcRejectCode', '>u2'),
            ('RejectParameter1', '>u2'),
            ('RejectParameter2', '>u2')]))
        print('[TcReject]: ', self, packet['TcPacketId'][0],
              packet['TcSeqControl'][0], bin(packet['TcRejectCode'][0]),
              packet['RejectParameter1'][0], packet['RejectParameter2'][0])
        return packet

    def __rd_other(self, hdr) -> np.ndarray:
        """
        Read other telemetry packet
        """
        num_bytes = self.packet_length - TIME_DTYPE.itemsize + 1
        if not 0x320 <= self.ap_id <= 0x350:
            self.found_invalid_apid = True
            self.fp.seek(num_bytes, 1)
            return None

        packet = np.empty(1, dtype=np.dtype([
            ('packet_header', HDR_DTYPE),
            ('raw_data', 'u1', num_bytes)]))
        packet['packet_header'] = hdr
        packet['raw_data'] = np.fromfile(self.fp, count=num_bytes, dtype='u1')
        return packet

    def read_packet(self) -> np.ndarray:
        """
        Read next telemetry packet

        Returns
        -------
        ndarray with packet data
        """
        # read primary/secondary header
        hdr = np.fromfile(self.fp, count=1, dtype=HDR_DTYPE)
        if hdr.size == 0:
            try:
                self.open_next_file()
            except StopIteration:
                return None
            except FileNotFoundError:
                return None

            hdr = np.fromfile(self.fp, count=1, dtype=HDR_DTYPE)
            if hdr.size == 0:
                return None

        # save packet header as class attribute
        self.__hdr = hdr[0]

        # read telemetry packet
        rd_func = {0x350: self.__rd_science,
                   0x320: self.__rd_nomhk,
                   0x322: self.__rd_demhk,
                   0x331: self.__rd_tc_accept,
                   0x332: self.__rd_tc_reject,
                   0x333: self.__rd_tc_execute,
                   0x334: self.__rd_tc_fail}.get(self.ap_id, self.__rd_other)
        return rd_func(hdr)

    @staticmethod
    def select_tm(packets_in: tuple, ap_id: int) -> tuple:
        """
        Select telemetry packages on SPEXone ApID

        Parameters
        ----------
        packets: tuple
           Tuple with mix of SPEXone telemetry packages
        ap_id: int
          SPEXone ApID

        Returns
        -------
        Tuple with selected telemetry packages
        """
        packets = ()
        for packet in packets_in:
            if 'packet_header' not in packet.dtype.names:
                continue

            if (packet['packet_header']['type'] & 0x7FF) == ap_id:
                packets += (packet,)

        return packets

    def science_tm(self, packets_in: tuple) -> tuple:
        """
        Combine segmented Science telemetry packages

        Parameters
        ----------
        packets: tuple
           Tuple with science or house-keeping telemetry packages

        Returns
        -------
        Tuple with unsegmented Science telemetry packages
        """
        # reject non-Science telemetry packages
        packets = self.select_tm(packets_in, 0x350)
        if not packets:
            return ()

        # check if grouping_flag of first segement equals 1
        #   else reject all segments with grouping_flag != 1
        self.__hdr = packets[0]['packet_header']
        if self.grouping_flag != 1:
            ii = 0
            for packet in packets:
                self.__hdr = packet['packet_header']
                if self.grouping_flag == 1:
                    break
                ii += 1

            print(f'[WARNING]: first frame incomplete - skipped {ii} segments')
            packets = packets[ii:]
            if not packets:
                return ()

        # check if grouping_flag of last segement equals 2
        #   else reject all segments after the last segment
        #   with grouping_flag == 2
        self.__hdr = packets[-1]['packet_header']
        if self.grouping_flag != 2:
            ii = 0
            for packet in packets:
                self.__hdr = packet['packet_header']
                if self.grouping_flag == 2:
                    break
                ii += 1

            print(f'[WARNING]: last frame incomplete - rejected {ii} segments')
            packets = packets[:-ii]
            if not packets:
                return ()

        res = ()
        offs = 0
        prev_grp_flag = 2
        for packet in packets:
            grouping_flag = (packet['packet_header']['sequence'] >> 14) & 0x3
            # print(prev_grp_flag, grouping_flag, len(res), offs,
            #      packet['image_data'].size)
            # handle segmented data
            if grouping_flag == 1:       # first segment
                # group_flag of previous package should be 2
                if prev_grp_flag != 2:
                    if packet['image_data'].size in (3853, 7853):
                        print(MSG_SKIP_FRAME)
                        prev_grp_flag = 2
                        offs = 0
                    else:
                        raise RuntimeError(MSG_CORRUPT_APID)

                img_size = packet['science_hk']['IMRLEN'] // 2
                rec_buff = np.empty(1, dtype=np.dtype([
                    ('packet_header', HDR_DTYPE),
                    ('science_hk', SCIHK_DTYPE),
                    ('icu_time', TIME_DTYPE),
                    ('image_data', 'O')]))[0]
                img_buff = np.empty(img_size, dtype='u2')

                rec_buff['packet_header'] = packet['packet_header']
                rec_buff['science_hk'] = packet['science_hk']
                rec_buff['icu_time'] = packet['icu_time']
                img_buff[offs:offs + packet['image_data'].size] = \
                    packet['image_data']
                offs += packet['image_data'].size
            else:                        # continuation or last segment
                # group_flag of previous package should be 0 or 1
                if prev_grp_flag == 2:
                    raise RuntimeError(MSG_CORRUPT_FRAME)

                img_buff[offs:offs + packet['image_data'].size] = \
                    packet['image_data']
                offs += packet['image_data'].size
                if grouping_flag == 2:
                    if offs == img_size:
                        rec_buff['image_data'] = img_buff
                        res += (rec_buff,)
                    else:
                        print(MSG_SKIP_FRAME)
                    offs = 0

            # keep current group flag for next read
            prev_grp_flag = grouping_flag

        return res



"""
This file is part of pyspex

https://github.com/rmvanhees/pyspex.git

Python class add EGSE/OGSE data to a SPEXone Level-1A product

Copyright (c) 2021 SRON - Netherlands Institute for Space Research
   All Rights Reserved

License:  BSD-3-Clause
"""
import numpy as np

from netCDF4 import Dataset


# --------------------------------------------------
class LV1gse:
    """
    Python class add EGSE/OGSE data to a SPEXone Level-1A product
    """
    def __init__(self, l1a_file: str) -> None:
        """
        Initialize netCDF4 group 'gse_data' in a SPEXone Level-1 product

        Attributes
        ----------
        fid : netCDF4::Dataset
            netCDF4 Dataset instance.

        Methods
        -------
        check_egse(egse_data)
           Check consistency of OGSE/EGSE information during measurement.
        set_attr(name, value)
           Add/update an attribute of the group 'gse_data'
        write_attr_act(angle: float, illumination)
           Add ACT rotation angle as an group attribute.
        write_attr_alt(angle: float, illumination)
           Add altitude rotation angle as an group attribute.
        write_attr_polarization(aolp: float, dolp: float)
           Add polarization parameters AoLP & DoLP as group attributes.
        write_viewport(viewport)
           Add/update which viewports are illuminated.
        write_egse(egse_time, egse_data, egse_attrs)
           Add EGSE parameters.
        write_data_stimulus(xds_signal)
           Add wavelength and signal of data stimulus.
        write_reference_diode(ref_time, ref_data, ref_attrs)
           Add data measured by the reference diode during the measurement.
        write_reference_signal(xds_signal)
           Add reference signal and its error.
        write_wavelength_monitor(xds_wavelength)
           Add wavelength monitoring data of the Avantas fibre-spectrometer.
        """
        self.fid = Dataset(l1a_file, 'r+')

        if self.fid.groups.get("gse_data"):
            return

        # investigate filename
        parts = l1a_file.split('_')
        if len(parts) > 2 and parts[0] == 'SPX1':
            parts = parts[2:]
        msmt_fields = parts[0].split('-')
        background = 'BKG' in msmt_fields
        act_angle = [float(x.replace('act', ''))
                     for x in parts if x.startswith('act')]
        alt_angle = [float(x.replace('alt', ''))
                     for x in parts if x.startswith('alt')]
        pol_angle = [float(x.replace('pol', ''))
                     for x in parts if x.startswith('pol') and x != 'polcal']
        gp1_angle = [float(x.replace('glass', ''))
                     for x in parts if x.startswith('glass')]
        gp1_offs = 5.634375
        # gp2_offs = 5.09625

        # determine viewport: default 0, when all viewports are illuminated
        if alt_angle:
            vp_angle = np.array([-50., -20, 0, 20, 50])
            vp_diff = np.abs(vp_angle - alt_angle[0])
            if vp_diff.min() > 6:
                viewport = 0
            else:
                viewport = 2 ** np.argmin(vp_diff)
        else:
            viewport = 0

        gid = self.fid.createGroup('/gse_data')
        dset = gid.createVariable('viewport', 'u1')
        dset.long_name = "viewport status"
        dset.standard_name = "status_flag"
        dset.valid_range = np.array([0, 16], dtype='u1')
        dset.flag_values = np.array([0, 1, 2, 4, 8, 16], dtype='u1')
        dset.flag_meanings = "ALL -50deg -20deg 0deg +20deg +50deg"
        dset[:] = viewport

        # gid.FOV_begin = np.nan
        # gid.FOV_end = np.nan
        gid.ACT_rotationAngle = np.nan if not act_angle else act_angle[0]
        gid.ALT_rotationAngle = np.nan if not alt_angle else alt_angle[0]
        # gid.ACT_illumination = np.nan
        # gid.ALT_illumination = np.nan
        gid.AoLP = 0. if not pol_angle else pol_angle[0]
        if not background and msmt_fields[0] in ('POLARIZED', 'POLARIMETRIC'):
            gid.DoLP = 1.
            if gp1_angle:
                gid.GP1_angle = gp1_angle[0] + gp1_offs
        else:
            gid.DoLP = 0.

    def __iter__(self) -> None:
        for attr in sorted(self.__dict__):
            if not attr.startswith("__"):
                yield attr

    def __enter__(self):
        """
        method called to initiate the context manager
        """
        return self

    def __exit__(self, exc_type, exc_value, traceback) -> bool:
        """
        method called when exiting the context manager
        """
        self.close()
        return False  # any exception is raised by the with statement.

    def close(self) -> None:
        """
        Close all resources (currently a placeholder function)
        """
        if self.fid is None:
            return

        self.fid.close()
        self.fid = None

    def check_egse(self, egse_data) -> None:
        """
        Check consistency of OGSE/EGSE information during measurement
        """
        for key, fmt in egse_data.dtype.fields.items():
            if fmt[0] == np.uint8:
                res_sanity = (egse_data[key] == egse_data[key][0]).all()
                if not res_sanity:
                    print('[WARNING] ', key, egse_data[key])

        act_angle = self.fid['/gse_data'].ACT_rotationAngle
        if np.isfinite(act_angle):
            if not np.allclose(egse_data['ACT_ANGLE'], act_angle, 1e-2):
                print('[WARNING] ', 'ACT_ANGLE', egse_data['ACT_ANGLE'])

        alt_angle = self.fid['/gse_data'].ALT_rotationAngle
        if np.isfinite(alt_angle):
            if not np.allclose(egse_data['ALT_ANGLE'], alt_angle, 1e-2):
                print('[WARNING] ', 'ALT_ANGLE', egse_data['ALT_ANGLE'])

    def set_attr(self, name: str, value) -> None:
        """
        Add attribute to group 'gse_data'

        Parameters
        ----------
        name: str
        value: anything(?)
        """
        self.fid['/gse_data'].setncattr(name, value)

    def write_attr_act(self, angle: float, illumination=None) -> None:
        """
        Add act rotation angle as an group attribute

        Parameters
        ----------
        angle :  float
        illumination :  float
        """
        self.fid['/gse_data'].ACT_rotationAngle = angle
        if illumination is not None:
            self.fid['/gse_data'].ACT_illumination = illumination

    def write_attr_alt(self, angle: float, illumination=None) -> None:
        """
        Add altitude rotation angle as an group attribute

        Parameters
        ----------
        angle :  float
        illumination :  float
        """
        self.fid['/gse_data'].ALT_rotationAngle = angle
        if illumination is not None:
            self.fid['/gse_data'].ALT_illumination = illumination

    def write_attr_polarization(self, aolp: float, dolp: float) -> None:
        """
        Add polarization parameters AoLP & DoLP as group attributes

        Parameters
        ----------
        AoLP :  float
           Angle of linear polarization
        DoLP :  float
           Degree of linear polarization
        """
        if aolp is not None:
            self.fid['/gse_data'].AoLP = aolp
        if dolp is not None:
            self.fid['/gse_data'].DoLP = dolp

    def write_egse(self, egse_time, egse_data, egse_attrs: dict) -> None:
        """
        Add EGSE parameters

        Parameters
        ----------
        egse_time :  ndarray
        egse_data :  ndarray
        egse_attrs :  dict
        """
        # perform sanity check on EGSE parameters
        self.check_egse(egse_data)

        gid = self.fid['/gse_data']
        _ = gid.createDimension('time', len(egse_data))
        dset = gid.createVariable('time', 'f8', ('time',))
        dset[:] = egse_time

        egse_t = gid.createCompoundType(egse_data.dtype, 'egse_dtype')
        dset = gid.createVariable('egse', egse_t, ('time',))
        dset.long_name = egse_attrs['long_name']
        dset.fields = egse_attrs['fields']
        dset.units = egse_attrs['units']
        dset.comment = egse_attrs['comment']
        dset[:] = egse_data

    def write_viewport(self, viewport: int) -> None:
        """
        Add/update which viewports are illuminated

        Parameters
        ----------
        viewport :  int
        """
        self.fid['/gse_data/viewport'][:] = viewport

    def write_reference_signal(self, signal, stdev) -> None:
        """
        Write reference detector signal and variance

        Parameters
        ----------
        signal :  float
           Mean of signal level measured during the measurement
        stdev :  float
           standard deviation of signal level measured during the measurement

        Notes
        -----
        Used for non-linearity measurements.
        """
        gid = self.fid['/gse_data']
        gid.Illumination_level = 5e9 / 1.602176634 * signal

        dset = gid.createVariable('reference_signal', 'f8', ())
        dset.long_name = "mean of reference-diode signal"
        dset.comment = "t_sat = min(2.28e-9 / S_reference, 30)"
        dset.units = 'A'
        dset[:] = signal

        dset = gid.createVariable('reference_signal_std', 'f8', ())
        dset.long_name = "standard deviation of reference-diode signal"
        dset.units = 'A'
        dset[:] = stdev

    def write_reference_diode(self, ref_time, ref_data, ref_attrs) -> None:
        """
        Add data measured by the reference diode during the measurement

        Parameters
        ----------
        ref_time :  ndarray
        ref_data :  ndarray
        ref_attrs :  dict
           fid['/ReferenceDiode/ref_diode'].__dict__
        """
        gid = self.fid.createGroup('/gse_data/ReferenceDiode')
        _ = gid.createDimension('time', len(ref_data))
        dset = gid.createVariable('time', 'f8', ('time',))
        dset.units = 'seconds since 1970-01-01T00:00:00+00:00'
        dset.comment = 'Generated on SRON clean-room tablet'
        dset[:] = ref_time

        ref_t = gid.createCompoundType(ref_data.dtype, 'ref_dtype')
        dset = gid.createVariable('ref_diode', ref_t, ('time',))
        dset.setncatts(ref_attrs)
        dset[:] = ref_data

    def write_wavelength_monitor(self, xds_wav_mon) -> None:
        """
        Add wavelength monitoring data of the Avantas fibre-spectrometer

        Parameters
        ----------
        xds_wav_mon :  xarray::Dataset
           Contains xarray::DataArrays 'signal', 'wavelength', 'wav_time'
        """
        xds_wav_mon.to_netcdf(self.fid.filepath(), mode='a',
                              group='/gse_data/WaveMonitor')

        # gid = self.fid.createGroup('/gse_data/WaveMonitor')
        # _ = gid.createDimension('time', len(wav_time))
        # dset = gid.createVariable('time', 'f8', ('time',))
        # dset.units = 'seconds since 1970-01-01T00:00:00+00:00'
        # dset.comment = 'Generated on SRON clean-room tablet'
        # dset[:] = wav_time
        # _ = gid.createDimension('wavelength', len(wav_wv))
        # dset = gid.createVariable('wavelength', 'f4', ('wavelength',))
        # dset.longname = 'wavelength grid'
        # dset.comment = 'wavelength annotation of fibre spectrometer'
        # dset[:] = wav_wv
        # dset = gid.createVariable('t_intg', 'i2', ('time',))
        # dset.long_name = 'Integration time'
        # dset.units = 'ms'
        # dset[:] = wav_intg
        # dset = gid.createVariable('n_avg', 'i2', ('time',))
        # dset.long_name = 'Averaging number'
        # dset[:] = wav_avg_num
        # dset = gid.createVariable('wav_mon', 'f4', ('time', 'wavelength'))
        # dset.long_name = 'wavelength-monitor spectra'
        # dset.comment = 'Avantes fibre spectrometer'
        # dset[:] = wav_signal



"""
This file is part of pyspex

https://github.com/rmvanhees/pyspex.git

Python implementation of the PACE SPEX Level-1B product

Copyright (c) 2020-2021 SRON - Netherlands Institute for Space Research
   All Rights Reserved

License:  BSD-3-Clause
"""
from netCDF4 import Dataset

# - global parameters ------------------------------


# - local functions --------------------------------


# - main function ----------------------------------
def init_l1b(l1b_flname: str, ref_date, dims: dict):
    """
    Create an empty PACE SPEX Level-1B product

    Parameters
    ----------
    l1b_flname : string
       Name of Level-1B product
    ref_date : datetime.datetime
       Reference date for image time
    dims :   dictionary
       Provide length of the Level-1B dimensions
       Default values:
          bins_along_track :  400
          spatial_samples_per_image :  200
          intensity_bands_per_view :  50
          polarization_bands_per_view :  50

    """
    # check function parameters
    if not isinstance(dims, dict):
        raise TypeError("dims should be a dictionary")

    # initialize dimensions
    n_views = 5
    n_bins_along = 400
    n_spatial_samples = 200
    n_intens_bands = 50
    n_polar_bands = 50

    if 'bins_along_track' in dims:
        n_bins_along = dims['bins_along_track']
    if 'spatial_samples_per_image' in dims:
        n_spatial_samples = dims['spatial_samples_per_image']
    if 'intensity_bands_per_view' in dims:
        n_intens_bands = dims['intensity_bands_per_view']
    if 'polarization_bands_per_view' in dims:
        n_polar_bands = dims['polarization_bands_per_view']

    # create/overwrite netCDF4 product
    rootgrp = Dataset(l1b_flname, "w")

    # create global dimensions
    _ = rootgrp.createDimension('number_of_views', n_views)
    _ = rootgrp.createDimension('bins_along_track', n_bins_along)
    _ = rootgrp.createDimension('spatial_samples_per_image', n_spatial_samples)
    _ = rootgrp.createDimension('intensity_bands_per_view', n_intens_bands)
    _ = rootgrp.createDimension('polarization_bands_per_view', n_polar_bands)

    # create groups and all variables with attributes
    sgrp = rootgrp.createGroup('BIN_ATTRIBUTES')
    chunksizes = None if n_bins_along is not None else (512,)
    dset = sgrp.createVariable('image_time', 'f8', ('bins_along_track',),
                               chunksizes=chunksizes)
    dset.long_name = 'image time (seconds of day)'
    dset.valid_min = 0
    dset.valid_max = 86400.999999
    if ref_date is None:
        dset.units = 'seconds'
    else:
        dset.units = f'seconds since {ref_date.isoformat(sep=" ")}'

    # -------------------------
    sgrp = rootgrp.createGroup('GEOLOCATION_DATA')
    chunksizes = (None if n_bins_along is not None
                  else (128, n_spatial_samples))
    dset = sgrp.createVariable('altitude', 'f4',
                               ('bins_along_track',
                                'spatial_samples_per_image'),
                               chunksizes=chunksizes)
    dset.standard_name = 'altitude'
    dset.long_name = "height above mean sea level"
    dset.units = "meters"
    dset.positive = "up"
    dset.axis = "Z"
    dset = sgrp.createVariable('latitude', 'f4',
                               ('bins_along_track',
                                'spatial_samples_per_image'),
                               chunksizes=chunksizes)
    dset.standard_name = 'latitude'
    dset.long_name = 'latitude'
    dset.valid_min = -90
    dset.valid_max = 90
    dset.units = 'degrees_north'
    dset = sgrp.createVariable('longitude', 'f4',
                               ('bins_along_track',
                                'spatial_samples_per_image'),
                               chunksizes=chunksizes)
    dset.standard_name = 'longitude'
    dset.long_name = 'longitude'
    dset.valid_min = -180
    dset.valid_max = 180
    dset.units = 'degrees_east'
    dset = sgrp.createVariable('sensor_azimuth', 'f4',
                               ('bins_along_track',
                                'spatial_samples_per_image'),
                               chunksizes=chunksizes)
    dset.long_name = 'sensor azimuth angle'
    dset.units = 'degrees'
    dset = sgrp.createVariable('sensor_zenith', 'f4',
                               ('bins_along_track',
                                'spatial_samples_per_image'),
                               chunksizes=chunksizes)
    dset.long_name = 'sensor zenith angle'
    dset.units = 'degrees'
    dset = sgrp.createVariable('solar_azimuth', 'f4',
                               ('bins_along_track',
                                'spatial_samples_per_image'),
                               chunksizes=chunksizes)
    dset.long_name = 'solar azimuth angle'
    dset.units = 'degrees'
    dset = sgrp.createVariable('solar_zenith', 'f4',
                               ('bins_along_track',
                                'spatial_samples_per_image'),
                               chunksizes=chunksizes)
    dset.long_name = 'solar zenith angle'
    dset.units = 'degrees'

    # -------------------------
    sgrp = rootgrp.createGroup('OBSERVATION_DATA')
    chunksizes = None if n_bins_along is not None else \
        (8, n_spatial_samples, n_intens_bands)
    dset = sgrp.createVariable('I', 'f4',
                               ('bins_along_track', 'spatial_samples_per_image',
                                'intensity_bands_per_view'),
                               chunksizes=chunksizes)
    dset.long_name = 'I Stokes vector component'
    dset.units = 'W.m-2.sr-1.um-1'
    dset = sgrp.createVariable('I_noise', 'f4',
                               ('bins_along_track', 'spatial_samples_per_image',
                                'intensity_bands_per_view'),
                               chunksizes=chunksizes)
    dset.long_name = 'random noise of I'
    dset.units = 'W.m-2.sr-1.um-1'

    chunksizes = None if n_bins_along is not None else \
        (64, n_spatial_samples, n_polar_bands)
    dset = sgrp.createVariable('Q_over_I', 'f4',
                               ('bins_along_track', 'spatial_samples_per_image',
                                'polarization_bands_per_view'),
                               chunksizes=chunksizes)
    dset.long_name = 'Q over I (little q) Stokes vector component'
    dset.units = '1'
    dset = sgrp.createVariable('Q_over_I_noise', 'f4',
                               ('bins_along_track', 'spatial_samples_per_image',
                                'polarization_bands_per_view'),
                               chunksizes=chunksizes)
    dset.long_name = 'random noise of Q_over_I'
    dset.units = '1'
    dset = sgrp.createVariable('U_over_I', 'f4',
                               ('bins_along_track', 'spatial_samples_per_image',
                                'polarization_bands_per_view'),
                               chunksizes=chunksizes)
    dset.long_name = 'U over I (little u) Stokes vector component'
    dset.units = '1'
    dset = sgrp.createVariable('U_over_I_noise', 'f4',
                               ('bins_along_track', 'spatial_samples_per_image',
                                'polarization_bands_per_view'),
                               chunksizes=chunksizes)
    dset.long_name = 'random noise of U_over_I'
    dset.units = '1'
    dset = sgrp.createVariable('AoLP', 'f4',
                               ('bins_along_track', 'spatial_samples_per_image',
                                'polarization_bands_per_view'),
                               chunksizes=chunksizes)
    dset.long_name = 'angle of linear polarization'
    dset.units = 'degrees'
    dset = sgrp.createVariable('AoLP_noise', 'f4',
                               ('bins_along_track', 'spatial_samples_per_image',
                                'polarization_bands_per_view'),
                               chunksizes=chunksizes)
    dset.long_name = 'random noise of AoLP'
    dset.units = 'degrees'
    dset = sgrp.createVariable('DoLP', 'f4',
                               ('bins_along_track', 'spatial_samples_per_image',
                                'polarization_bands_per_view'),
                               chunksizes=chunksizes)
    dset.long_name = 'degree of linear polarization'
    dset.units = '1'
    dset = sgrp.createVariable('DoLP_noise', 'f4',
                               ('bins_along_track', 'spatial_samples_per_image',
                                'polarization_bands_per_view'),
                               chunksizes=chunksizes)
    dset.long_name = 'random noise of DoLP'
    dset.units = '1'

    # -------------------------
    sgrp = rootgrp.createGroup('SENSOR_VIEWS_BANDS')
    dset = sgrp.createVariable("viewport_index", "u1",
                               ("spatial_samples_per_image",))
    dset.long_name = "index of viewport"
    dset.valid_min = 0
    dset.valid_max = 4
    dset.comment = "Contains indices to each viewport for all spatial samples."
    dset = sgrp.createVariable('view_angles', 'f4', ('number_of_views',))
    dset.long_name = 'along track view zenith angles at sensor'
    dset.units = 'degrees'
    dset.comment = ('view_angles is defined at the sensor, as it provides'
                    ' a swath independent value at TOA.')
    dset = sgrp.createVariable('intensity_wavelengths', 'f4',
                               ('number_of_views', 'intensity_bands_per_view'))
    dset.long_name = 'wavelength at center of intensity bands'
    dset.units = 'nm'
    dset = sgrp.createVariable('intensity_bandpasses', 'f4',
                               ('number_of_views', 'intensity_bands_per_view'))
    dset.long_name = 'FWHM of intensity bands'
    dset.units = 'nm'
    dset = sgrp.createVariable('intensity_F0', 'f4',
                               ('number_of_views', 'intensity_bands_per_view'))
    dset.long_name = 'Solar irradiance on intensity wavelength grid'
    dset.units = 'W.m-2'
    dset.comment = ('Spectral response function convolved mean solar flux'
                    ' at each intensity band and view.')
    dset = sgrp.createVariable('polarization_wavelengths', 'f4',
                               ('number_of_views',
                                'polarization_bands_per_view'))
    dset.long_name = 'wavelength at center of polarization bands'
    dset.units = 'nm'
    dset = sgrp.createVariable('polarization_bandpasses', 'f4',
                               ('number_of_views',
                                'polarization_bands_per_view'))
    dset.long_name = 'FWHM of polarization bands'
    dset.units = 'nm'
    dset = sgrp.createVariable('polarization_F0', 'f4',
                               ('number_of_views',
                                'polarization_bands_per_view'))
    dset.long_name = 'Solar irradiance on polarization wavelength grid'
    dset.units = 'W.m-2'
    dset.comment = ('Spectral response function convolved mean solar flux'
                    ' at each polarization band and view.')

    return rootgrp


# --------------------------------------------------
if __name__ == '__main__':
    fid = init_l1b('PACE_SPEX.20230115T123456.L1B.2.5km.V01.nc',
                   ref_date=None, dims={})
    fid.fill_global_attrs(orbit=12345, bin_size='2.5km')
    fid.close()



#!/usr/bin/env python3
"""
This file is part of pyspex

https://github.com/rmvanhees/pyspex.git

Python script to store SPEXone Level-0 data in a new Level-1A product.

Examples
--------
Read ST3 Level-0 file in current directory, write Level-1A in different
directory:

   spx1_level01a.py --datapath L1A ./SCI_20220124_174737_419.ST3

Read CCSDS Level-0 files in current directory, write Level-1A in different
directory:

   spx1_level01a.py --datapath L1A ./NomSciCal1_20220123T121801.676167

Note that science data is read from ./NomSciCal1_20220123T121801.676167.?
and ./NomSciCal1_20220123T121801.676167.??, and telemetry data is read
from ./NomSciCal1_20220123T121801.676167_hk.?

Read ST3 Level-0 file in current directory, and write packet header information
to a file with extension '.dump' in directory 'L1A':

   spx1_level01a.py --datapath L1A ./SCI_20220124_174737_419.ST3 --dump

Note that the extension will not be removed, thus the name of the dump file is:
./L1A/SCI_20220124_174737_419.ST3.dump

Copyright (c) 2022 SRON - Netherlands Institute for Space Research
   All Rights Reserved

License:  BSD-3-Clause
"""
import argparse

from datetime import datetime, timedelta, timezone
from pathlib import Path

import numpy as np

#from pyspex import spx_product
from pyspex.lv1_io import L1Aio
from pyspex.st3_io import (ap_id, split_ccsds, select_science, select_hk,
                           ScienceCCSDS, TmTcCCSDS)

# - global parameters ------------------------------
EPOCH_1958 = datetime(1958, 1, 1, tzinfo=timezone.utc)
EPOCH_1970 = datetime(1970, 1, 1, tzinfo=timezone.utc)


# - local functions --------------------------------
def __rd_l0_data(args) -> tuple:
    """
    Read level 0 data and return Science and telemetry data
    """
    # concatenate CCSDS Science data of the input files
    if args.msmt_id.name.endswith('.ST3'):
        with open(args.msmt_id, 'rb') as fp:
            ccsds_data, ccsds_hk = split_ccsds(fp.read())
    else:
        data_dir = args.msmt_id.parent
        sci_files = sorted(data_dir.glob(args.msmt_id.name + '.[0-9]')) \
            + sorted(data_dir.glob(args.msmt_id.name + '.?[0-9]'))
        ccsds_data = ()
        for flname in sci_files:
            with open(flname, 'rb') as fp:
                ccsds_data += select_science(fp.read())

        hk_files = sorted(data_dir.glob(args.msmt_id.name + '_hk.[0-9]'))
        print(hk_files)
        ccsds_hk = ()
        for flname in hk_files:
            with open(flname, 'rb') as fp:
                ccsds_hk += select_hk(fp.read())

    # perform data dump and exit
    if args.dump:
        # dump header information of the Science packages
        ScienceCCSDS(ccsds_data, args.verbose).dump(
            args.datapath / (args.msmt_id.name + '.dump'))
        # dump header information of the Science packages
        TmTcCCSDS(ccsds_hk, args.verbose).dump(
            args.datapath / (args.msmt_id.name + '_hk.dump'))
        return None

    # read Science packages and collect all detector read-outs
    sci_list = ScienceCCSDS(ccsds_data, args.verbose).read(args.select)
    if not sci_list:
        print('[WARNING]: no science data found or selected')
        return None

    science = np.concatenate(sci_list)
    mps_list = np.unique(science['hk']['MPS_ID']).tolist()
    if args.verbose:
        print(f'[INFO]: list of unique MPS {mps_list}')

    tmtc_list = TmTcCCSDS(ccsds_hk, args.verbose).read()
    nomhk = np.concatenate(
        [(x,) for x in tmtc_list if ap_id(x['hdr']) == 0x320
         and x['hk']['MPS_ID'] in mps_list])

    return (science, nomhk)


def get_science_timestamps(science):
    """
    Return timestamps of the Science packets
    """
    if science['hk']['ICUSWVER'][0] > 0x123:
        img_sec = science['icu_tm']['tai_sec']
        img_subsec = science['icu_tm']['sub_sec']
        return (img_sec, img_subsec)

    img_sec = science['hdr']['tai_sec']
    img_subsec = science['hdr']['sub_sec']
    if science['hk']['ICUSWVER'][0] > 0x123:
        # fix bug in sub-seconds
        us100 = np.round(10000 * img_subsec.astype(float) / 65536)
        buff = us100 + img_sec - 10000
        us100 = buff.astype('u8') % 10000
        img_subsec = ((us100 << 16) // 10000).astype('u2')
    return (img_sec, img_subsec)


def get_nomhk_timestamps(nomhk):
    """
    Return timestamps of the telemetry packets
    """
    nomhk_sec = nomhk['hdr']['tai_sec']
    nomhk_subsec = nomhk['hdr']['sub_sec']
    if nomhk['hk']['ICUSWVER'][0] > 0x123:
        # fix bug in sub-seconds
        us100 = np.round(10000 * nomhk_subsec.astype(float) / 65536)
        buff = us100 + nomhk_sec - 10000
        us100 = buff.astype('u8') % 10000
        nomhk_subsec = ((us100 << 16) // 10000).astype('u2')
    return (nomhk_sec, nomhk_subsec)


def get_l1a_name(args, science) -> str:
    """
    Generate name of Level-1A product, using the following filename conventions

    Inflight
    --------
    L1A file name format, following the NASA ... naming convention:
       PACE_SPEXone[_TTT].YYYYMMDDTHHMMSS.L1A.Vnn.nc
    where
       TTT is an optional data type (e.g., for the calibration data files)
       YYYYMMDDTHHMMSS is time stamp of the first image in the file
       nn file-version number
    for example
    [Science Product] PACE_SPEXone.20230115T123456.L1A.V01.nc
    [Calibration Product] PACE_SPEXone_CAL.20230115T123456.L1A.V01.nc
    [Monitoring Products] PACE_SPEXone_DARK.20230115T123456.L1A.V01.nc

    OCAL
    ----
    L1A file name format:
       SPX1_OCAL_<msm_id>_L1A_YYYYMMDDTHHMMSS_yyyymmddThhmmss_vvvv.nc
    where
       msm_id is the measurement identifier
       YYYYMMDDTHHMMSS is time stamp of the first image in the file
       yyyymmddThhmmss is the creation time (UTC) of the product
       vvvv is the version number of the product starting at 0001
    """
    img_sec, _ = get_science_timestamps(science)
    if args.msmt_id.name.endswith('.ST3'):
        # inflight product name
        # ToDo: detect Diagnostic DARK measurements
        prod_type = '_CAL' if args.select == 'fullFrame' else ''
        #sensing_start = EPOCH_1958 + timedelta(seconds=int(img_sec[0]))
        sensing_start = EPOCH_1970 + timedelta(seconds=int(img_sec[0]))

        return (f'PACE_SPEXone{prod_type}'
                f'.{sensing_start.strftime("%Y%m%dT%H%M%S"):15s}.L1A'
                f'.V{args.file_version:02d}.nc')

    # OCAL product name
    sensing_start = EPOCH_1970 + timedelta(seconds=int(img_sec[0]))

    # determine measurement identifier
    msm_id = args.msmt_id.name
    try:
        new_date = datetime.strptime(
            msm_id[-22:], '%y-%j-%H:%M:%S.%f').strftime('%Y%m%dT%H%M%S.%f')
    except ValueError:
        pass
    else:
        msm_id = msm_id[:-22] + new_date

    return (f'SPX1_OCAL_{msm_id}_L1A'
            f'_{sensing_start.strftime("%Y%m%dT%H%M%S"):15s}'
            f'_{datetime.utcnow().strftime("%Y%m%dT%H%M%S"):15s}'
            f'_{args.file_version:04d}.nc')


# - main function ----------------------------------
def main():
    """
    main function
    """
    # parse command-line parameters
    parser = argparse.ArgumentParser(
        formatter_class=argparse.RawTextHelpFormatter,
        description='store Level-0 data in a new SPEXone Level-1A product')
    parser.add_argument('--dump', action='store_true',
                        help=('dump CCSDS packet headers in ASCII'))
    parser.add_argument('--verbose', action='store_true', help='be verbose')
    parser.add_argument('--file_version', type=int, default=1,
                        help='provide file version number of level-1A product')
    parser.add_argument('--select', default='all',
                        choices=['all', 'binned', 'fullFrame'],
                        help=('read "all" data (default) or select'
                              ' "binned" / "full-frame" detector-readouts'))
    parser.add_argument('--datapath', type=Path, default=Path('.'),
                        help='directory to store the Level-1A product')
    # parser.add_argument('--st3_nav', default=None, type=str,
    #                    help='name of ST3 file with navigation data')
    # Note that science packages and telementry packages are combined in one
    # ST3 product (in chronological order), but seperated in CCSDS products.
    parser.add_argument('msmt_id', type=Path,
                        help=('[ST3] full name of the SPEXone level 0 file.'
                              '[CCSDS] filename (without extension) of the'
                              ' measurement data, where the expected filenames'
                              ' of the telemetry data are msmt_id + "_hk.?"'))
    args = parser.parse_args()
    if args.verbose:
        print(args)

    # read level 0 data
    if (res := __rd_l0_data(args)) is None:
        return

    science, nomhk = res
    if args.verbose:
        print(f'[INFO]: number of Science packages = {science.size}')
        print(f'[INFO]: number of telemetry packages = {nomhk.size}')
    dims = {'number_of_images': science.size,
            'samples_per_image': science['hk']['IMRLEN'].max() // 2,
            'hk_packets': nomhk.size,
            'SC_records': None}
    print(dims)

    # generate name of the level-1A product
    prod_name = get_l1a_name(args, science)

    # Generate and fill L1A product
    with L1Aio(args.datapath / prod_name, dims=dims) as l1a:
        # write image data, detector telemetry and image attributes
        img_data = np.empty((science.size, dims['samples_per_image']),
                            dtype=float)
        for ii, data in enumerate(science['frame']):
            img_data[ii, :data.size] = data
        l1a.fill_science(img_data, science['hk'],
                         np.bitwise_and(science['hdr']['sequence'], 0x3fff))
        del img_data
        img_sec, img_subsec = get_science_timestamps(science)
        l1a.fill_time(img_sec, img_subsec, group='image_attributes')

        # write engineering data
        if nomhk.size > 0:
            l1a.fill_nomhk(nomhk['hk'])
            nomhk_sec, nomhk_subsec = get_nomhk_timestamps(nomhk)
            l1a.fill_time(nomhk_sec, nomhk_subsec, group='engineering_data')

        # if demhk.size > 0:
        #    l1a.fill_demhk(demhk['hk'])

        # write global attributes
        l1a.fill_global_attrs(inflight=args.msmt_id.name.endswith('.ST3'))
        l1a.set_attr('input_files', str(args.msmt_id))


# --------------------------------------------------
if __name__ == '__main__':
    main()
